# -*- coding: utf-8 -*-
import asyncio
import unittest

import plumpy
from plumpy import BundleKeys

from . import utils


class TestWaitingProcess(unittest.TestCase):

    def test_instance_state(self):
        proc = utils.ThreeSteps()
        wl = utils.ProcessSaver(proc)
        proc.execute()

        for bundle, outputs in zip(wl.snapshots, wl.outputs):
            self.assertEqual(outputs, bundle.get(BundleKeys.OUTPUTS, {}))

    def test_saving_each_step(self):
        loop = asyncio.get_event_loop()
        for proc_class in utils.TEST_WAITING_PROCESSES:
            proc = proc_class()
            saver = utils.ProcessSaver(proc)
            saver.capture()

            self.assertTrue(utils.check_process_against_snapshots(loop, proc_class, saver.snapshots))

    def test_kill(self):
        process = utils.WaitForSignalProcess()

        # Kill the process when it enters the WAITING state
        listener = plumpy.ProcessListener()
        listener.on_process_waiting = lambda _proc: process.kill()
        process.add_process_listener(listener)

        with self.assertRaises(plumpy.KilledError):
            process.execute()
        self.assertTrue(process.killed())



# -*- coding: utf-8 -*-
"""
Keep track of the per-thread call stack of processes.
"""

import contextlib
import threading

# Use thread-local storage for the stack
THREAD_LOCAL = threading.local()


@contextlib.contextmanager
def in_stack(process):
    push(process)
    try:
        yield
    finally:
        pop(process)


def top():
    return _stack()[-1]


def stack():
    """Get an immutable tuple of the stack"""
    return tuple(_stack())


def push(process):
    _stack().append(process)


def is_empty():
    return len(_stack()) == 0


def pop(process):
    """
    Pop a process from the stack.  To make sure the stack is not corrupted
    the process instance of the calling process should be supplied
    so we can verify that is really is top of the stack.

    :param process: The process instance
    """
    global THREAD_LOCAL
    assert process is top(), "Can't pop a process that is not top of the stack"
    _stack().pop()


def _stack():
    """Access the private live stack"""
    global THREAD_LOCAL
    try:
        return THREAD_LOCAL.wf_stack
    except AttributeError:
        THREAD_LOCAL.wf_stack = []
        return THREAD_LOCAL.wf_stack



# -*- coding: utf-8 -*-
import time
import unittest

from plumpy.base import state_machine

# Events
PLAY = 'Play'
PAUSE = 'Pause'
STOP = 'Stop'

# States
PLAYING = 'Playing'
PAUSED = 'Paused'
STOPPED = 'Stopped'


class Playing(state_machine.State):
    LABEL = PLAYING
    ALLOWED = {PAUSED, STOPPED}
    TRANSITIONS = {STOP: STOPPED}

    def __init__(self, player, track):
        assert track is not None, 'Must provide a track name'
        super().__init__(player)
        self.track = track
        self._last_time = None
        self._played = 0.

    def __str__(self):
        if self.in_state:
            self._update_time()
        return '> {} ({}s)'.format(self.track, self._played)

    def enter(self):
        super().enter()
        self._last_time = time.time()

    def exit(self):
        super().exit()
        self._update_time()

    def play(self, track=None):  # pylint: disable=no-self-use, unused-argument
        return False

    def _update_time(self):
        current_time = time.time()
        self._played += current_time - self._last_time
        self._last_time = current_time


class Paused(state_machine.State):
    LABEL = PAUSED
    ALLOWED = {PLAYING, STOPPED}
    TRANSITIONS = {STOP: STOPPED}

    def __init__(self, player, playing_state):
        assert isinstance(playing_state, Playing), \
            'Must provide the playing state to pause'
        super().__init__(player)
        self.playing_state = playing_state

    def __str__(self):
        return '|| ({})'.format(self.playing_state)

    def play(self, track=None):
        if track is not None:
            self.state_machine.transition_to(Playing, track)
        else:
            self.state_machine.transition_to(self.playing_state)


class Stopped(state_machine.State):
    LABEL = STOPPED
    ALLOWED = {
        PLAYING,
    }
    TRANSITIONS = {PLAY: PLAYING}

    def __str__(self):
        return '[]'

    def play(self, track):
        self.state_machine.transition_to(Playing, track)


class CdPlayer(state_machine.StateMachine):
    STATES = (Stopped, Playing, Paused)

    def __init__(self):
        super().__init__()
        self.add_state_event_callback(
            state_machine.StateEventHook.ENTERING_STATE, lambda _s, _h, state: self.entering(state)
        )
        self.add_state_event_callback(state_machine.StateEventHook.EXITING_STATE, lambda _s, _h, _st: self.exiting())

    def entering(self, state):
        print('Entering {}'.format(state))
        print(self._state)

    def exiting(self):
        print('Exiting {}'.format(self.state))
        print(self._state)

    @state_machine.event(to_states=Playing)
    def play(self, track=None):
        return self._state.play(track)

    @state_machine.event(from_states=Playing, to_states=Paused)
    def pause(self):
        self.transition_to(Paused, self._state)
        return True

    @state_machine.event(from_states=(Playing, Paused), to_states=Stopped)
    def stop(self):
        self.transition_to(Stopped)


class TestStateMachine(unittest.TestCase):

    def test_basic(self):
        cd_player = CdPlayer()
        self.assertEqual(cd_player.state, STOPPED)

        cd_player.play('Eminem - The Real Slim Shady')
        self.assertEqual(cd_player.state, PLAYING)
        time.sleep(1.)

        cd_player.pause()
        self.assertEqual(cd_player.state, PAUSED)

        cd_player.play()
        self.assertEqual(cd_player.state, PLAYING)

        self.assertEqual(cd_player.play(), False)

        cd_player.stop()
        self.assertEqual(cd_player.state, STOPPED)

    def test_invalid_event(self):
        cd_player = CdPlayer()
        with self.assertRaises(AssertionError):
            cd_player.play()



# -*- coding: utf-8 -*-
"""
Python language utilities and tools.
"""
import functools
import inspect


def protected(check=False):

    def wrap(func):
        if isinstance(func, property):
            raise RuntimeError('Protected must go after @property decorator')

        args = inspect.getfullargspec(func)[0]
        if len(args) == 0:
            raise RuntimeError('Can only use the protected decorator on member functions')

        # We can only perform checks if the interpreter is capable of giving
        # us the stack i.e. currentframe() produces a valid object
        if check and inspect.currentframe() is not None:

            @functools.wraps(func)
            def wrapped_fn(self, *args, **kwargs):
                try:
                    calling_class = inspect.stack()[1][0].f_locals['self']
                    assert self is calling_class
                except (KeyError, AssertionError):
                    raise RuntimeError(
                        'Cannot access protected function {} from outside'
                        ' class hierarchy'.format(func.__name__)
                    )

                return func(self, *args, **kwargs)
        else:
            wrapped_fn = func

        return wrapped_fn

    return wrap


def override(check=False):

    def wrap(func):
        if isinstance(func, property):
            raise RuntimeError('Override must go after @property decorator')

        args = inspect.getfullargspec(func)[0]
        if len(args) == 0:
            raise RuntimeError('Can only use the override decorator on member functions')

        if check:

            @functools.wraps(func)
            def wrapped_fn(self, *args, **kwargs):
                try:
                    getattr(super(self.__class__, self), func.__name__)
                except AttributeError:
                    raise RuntimeError('Function {} does not override a superclass method'.format(func))

                return func(self, *args, **kwargs)
        else:
            wrapped_fn = func

        return wrapped_fn

    return wrap


class __NULL:  # pylint: disable=invalid-name

    def __eq__(self, other):
        return isinstance(other, self.__class__)


NULL = __NULL()



# -*- coding: utf-8 -*-
"""
Module containing future related methods and classes
"""
import asyncio
from typing import Coroutine

import kiwipy

__all__ = ['Future', 'gather', 'chain', 'copy_future', 'CancelledError', 'create_task']

CancelledError = kiwipy.CancelledError


class InvalidStateError(Exception):
    """Exception for when a future or action is in an invalid state"""


copy_future = kiwipy.copy_future  # pylint: disable=invalid-name
chain = kiwipy.chain  # pylint: disable=invalid-name
gather = asyncio.gather  # pylint: disable=invalid-name

Future = asyncio.Future  # pylint: disable=invalid-name


class CancellableAction(Future):
    """
    An action that can be launched and potentially cancelled
    """

    def __init__(self, action, cookie=None):
        super().__init__()
        self._action = action
        self._cookie = cookie

    @property
    def cookie(self):
        """ A cookie that can be used to correlate the actions with something """
        return self._cookie

    def run(self, *args, **kwargs):
        """
        Runt he action
        :param args: the positional arguments to the action
        :param kwargs: the keyword arguments to the action
        """
        if self.done():
            raise InvalidStateError('Action has already been ran')

        try:
            with kiwipy.capture_exceptions(self):
                self.set_result(self._action(*args, **kwargs))
        finally:
            self._action = None


def create_task(coro: Coroutine, loop=None):
    """
    Schedule a call to a coro in the event loop and wrap the outcome
    in a future.
    :param coro: the coro to schedule
    :param loop: the event loop to schedule it in
    :return: the future representing the outcome of the coroutine
    :rtype: :class:`plumpy.Future`
    """
    loop = loop or asyncio.get_event_loop()

    future = loop.create_future()

    async def run_task():
        with kiwipy.capture_exceptions(future):
            res = await coro()
            future.set_result(res)

    asyncio.run_coroutine_threadsafe(run_task(), loop)
    return future


def unwrap_kiwi_future(future):
    """
    Create a kiwi future that represents the final results of a nested series of futures,
    meaning that if the futures provided itself resolves to a future the returned
    future will not resolve to a value until the final chain of futures is not a future
    but a concrete value.  If at any point in the chain a future resolves to an exception
    then the returned future will also resolve to that exception.
    :param future: the future to unwrap
    :type future: :class:`kiwipy.Future`
    :return: the unwrapping future
    :rtype: :class:`kiwipy.Future`
    """
    unwrapping = kiwipy.Future()

    def unwrap(fut):
        if fut.cancelled():
            unwrapping.cancel()
        else:
            with kiwipy.capture_exceptions(unwrapping):
                result = fut.result()
                if isinstance(result, kiwipy.Future):
                    result.add_done_callback(unwrap)
                else:
                    unwrapping.set_result(result)

    future.add_done_callback(unwrap)
    return unwrapping



# -*- coding: utf-8 -*-
"""Utilities for tests"""
import asyncio
import collections
import shortuuid
import unittest

import kiwipy.rmq

import plumpy
from plumpy import processes, process_states, persistence, utils

Snapshot = collections.namedtuple('Snapshot', ['state', 'bundle', 'outputs'])


class TestCase(unittest.TestCase):
    pass


class DummyProcess(processes.Process):
    """
    Process with no inputs or outputs and does nothing when ran.
    """

    EXPECTED_STATE_SEQUENCE = [
        process_states.ProcessState.CREATED, process_states.ProcessState.RUNNING, process_states.ProcessState.FINISHED
    ]

    EXPECTED_OUTPUTS = {}


class DummyProcessWithOutput(processes.Process):
    EXPECTED_OUTPUTS = {'default': 5}

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.inputs.dynamic = True
        spec.outputs.dynamic = True
        spec.output('default', valid_type=int)

    def run(self, **kwargs):
        self.out('default', 5)


class DummyProcessWithDynamicOutput(processes.Process):
    EXPECTED_OUTPUTS = {'default': 5}

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.inputs.dynamic = True
        spec.outputs.dynamic = True

    def run(self, **kwargs):
        self.out('default', 5)


class KeyboardInterruptProc(processes.Process):

    @utils.override
    def run(self):
        raise KeyboardInterrupt()


class ProcessWithCheckpoint(processes.Process):

    @utils.override
    def run(self):
        return process_states.Continue(self.last_step)

    def last_step(self):
        pass


class WaitForSignalProcess(processes.Process):

    @utils.override
    def run(self):
        return process_states.Wait(self.last_step)

    def last_step(self):
        pass


class MissingOutputProcess(processes.Process):
    """ A process that does not generate a required output """

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.output('default', required=True)


class NewLoopProcess(processes.Process):

    def __init__(self, *args, **kwargs):
        kwargs['loop'] = plumpy.new_event_loop()
        super().__init__(*args, **kwargs)


class EventsTesterMixin:
    EVENTS = ('create', 'run', 'finish', 'emitted', 'wait', 'resume', 'stop', 'terminate')

    called_events = []

    @classmethod
    def called(cls, event):
        assert event in cls.EVENTS, "Unknown event '{}'".format(event)
        cls.called_events.append(event)

    def __init__(self, *args, **kwargs):
        assert isinstance(self, processes.Process), \
            'Mixin has to be used with a type derived from a Process'
        super().__init__(*args, **kwargs)
        self.__class__.called_events = []

    @utils.override
    def on_create(self):
        super().on_create()
        self.called('create')

    @utils.override
    def on_run(self):
        super().on_run()
        self.called('run')

    @utils.override
    def _on_output_emitted(self, output_port, value, dynamic):
        super()._on_output_emitted(output_port, value, dynamic)
        self.called('emitted')

    @utils.override
    def on_wait(self, wait_on):
        super().on_wait(wait_on)
        self.called('wait')

    @utils.override
    def on_resume(self):
        super().on_resume()
        self.called('resume')

    @utils.override
    def on_finish(self, result, successful):
        super().on_finish(result, successful)
        self.called('finish')

    @utils.override
    def on_stop(self):
        super().on_stop()
        self.called('stop')

    @utils.override
    def on_terminate(self):
        super().on_terminate()
        self.called('terminate')


class ProcessEventsTester(EventsTesterMixin, processes.Process):

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.outputs.dynamic = True

    def run(self):
        self.out('test', 5)


class ThreeSteps(ProcessEventsTester):
    EXPECTED_OUTPUTS = {'test': 5}

    _last_checkpoint = None

    @utils.override
    def run(self):
        self.out('test', 5)
        return process_states.Continue(self.middle_step)

    def middle_step(self,):
        return process_states.Continue(self.last_step)

    def last_step(self):
        pass


class TwoCheckpointNoFinish(ProcessEventsTester):

    def run(self):
        self.out('test', 5)
        return process_states.Continue(self.middle_step)

    def middle_step(self):
        pass


class ExceptionProcess(ProcessEventsTester):

    def run(self):
        self.out('test', 5)
        raise RuntimeError('Great scott!')


class ThreeStepsThenException(ThreeSteps):

    @utils.override
    def last_step(self):
        raise RuntimeError('Great scott!')


class ProcessListenerTester(plumpy.ProcessListener):

    def __init__(self, process, expected_events):
        process.add_process_listener(self)
        self.expected_events = set(expected_events)
        self.called = set()

    def on_process_created(self, process):
        self.called.add('created')

    def on_process_running(self, process):
        self.called.add('running')

    def on_process_waiting(self, process):
        self.called.add('waiting')

    def on_process_paused(self, process):
        self.called.add('paused')

    def on_output_emitted(self, process, output_port, value, dynamic):
        self.called.add('output_emitted')

    def on_process_finished(self, process, outputs):
        self.called.add('finished')

    def on_process_excepted(self, process, reason):
        self.called.add('excepted')

    def on_process_killed(self, process, msg):
        self.called.add('killed')


class Saver:

    def __init__(self):
        self.snapshots = []
        self.outputs = []

    def _save(self, p):
        b = persistence.Bundle(p)
        self.snapshots.append(b)
        self.outputs.append(p.outputs.copy())


class ProcessSaver(plumpy.ProcessListener, Saver):
    """
    Save the instance state of a process each time it is about to enter a new state
    """

    def __init__(self, proc):
        plumpy.ProcessListener.__init__(self)
        Saver.__init__(self)
        self.process = proc
        proc.add_process_listener(self)

    def capture(self):
        self._save(self.process)
        if not self.process.done():
            try:
                self.process.execute()
            except Exception:
                pass

    @utils.override
    def on_process_running(self, process):
        self._save(process)

    @utils.override
    def on_process_waiting(self, process, data):
        self._save(process)

    @utils.override
    def on_process_paused(self, process):
        self._save(process)

    # Terminal states:

    @utils.override
    def on_process_finished(self, process, outputs):
        self._save(process)

    @utils.override
    def on_process_excepted(self, process, reason):
        self._save(process)

    @utils.override
    def on_process_killed(self, process, msg):
        self._save(process)


# All the Processes that can be used
TEST_PROCESSES = [DummyProcess, DummyProcessWithOutput, DummyProcessWithDynamicOutput, ThreeSteps]

TEST_WAITING_PROCESSES = [
    ProcessWithCheckpoint, TwoCheckpointNoFinish, ExceptionProcess, ProcessEventsTester, ThreeStepsThenException
]

TEST_EXCEPTION_PROCESSES = [ExceptionProcess, ThreeStepsThenException, MissingOutputProcess]


def check_process_against_snapshots(loop, proc_class, snapshots):
    """
    Take the series of snapshots from a Process that executed and run it
    forward from each one.  Check that the subsequent snapshots match.
    This will only check up to the STARTED state because from that state back
    they should of course differ.

    Return True if they match, False otherwise.

    :param loop: The event loop
    :param proc_class: The process class to check
    :type proc_class: :class:`Process`
    :param snapshots: The snapshots taken from from an execution of that
      process
    :return: True if snapshots match False otherwise
    :rtype: bool
    """
    for i, bundle in zip(list(range(0, len(snapshots))), snapshots):
        loaded = bundle.unbundle(plumpy.LoadSaveContext(loop=loop))
        saver = ProcessSaver(loaded)
        saver.capture()

        # Now check going backwards until running that the saved states match
        j = 1
        while True:
            if j >= min(len(snapshots), len(saver.snapshots)):
                break

            compare_dictionaries(
                snapshots[-j], saver.snapshots[-j], snapshots[-j], saver.snapshots[-j], exclude={'exception'}
            )
            j += 1

    return True


def compare_dictionaries(bundle1, bundle2, dict1, dict2, exclude=None):
    keys = set(dict1.keys()) & set(dict2.keys())
    if exclude is not None:
        keys -= exclude

    for key in keys:
        if key not in dict1:
            raise ValueError("Key '{}' in dict 1 but not 2".format(key))

        if key not in dict2:
            raise ValueError("Key '{}' in dict 2 but not 1".format(key))

        v1 = dict1[key]
        v2 = dict2[key]

        compare_value(bundle1, bundle2, v1, v2, exclude)


def compare_value(bundle1, bundle2, v1, v2, exclude=None):
    if isinstance(v1, collections.Mapping) and isinstance(v2, collections.Mapping):
        compare_dictionaries(bundle1, bundle2, v1, v2, exclude)
    elif isinstance(v1, list) and isinstance(v2, list):
        for vv1, vv2 in zip(v1, v2):
            compare_value(bundle1, bundle2, vv1, vv2, exclude)
    else:
        if v1 != v2:
            raise ValueError('Dict values mismatch for :\n{} != {}'.format(v1, v2))


class TestPersister(persistence.Persister):
    """
    Test persister, just creates the bundle, noting else
    """

    def save_checkpoint(self, process, tag=None):
        """ Create the checkpoint bundle """
        persistence.Bundle(process)

    def load_checkpoint(self, pid, tag=None):
        raise NotImplementedError

    def get_checkpoints(self):
        raise NotImplementedError

    def get_process_checkpoints(self, pid):
        raise NotImplementedError

    def delete_checkpoint(self, pid, tag=None):
        raise NotImplementedError

    def delete_process_checkpoints(self, pid):
        raise NotImplementedError


def run_until_waiting(proc):
    """ Set up a future that will be resolved on entering the WAITING state """
    from plumpy import ProcessState

    listener = plumpy.ProcessListener()
    in_waiting = plumpy.Future()

    if proc.state == ProcessState.WAITING:
        in_waiting.set_result(True)
    else:

        def on_waiting(_waiting_proc):
            in_waiting.set_result(True)
            proc.remove_process_listener(listener)

        listener.on_process_waiting = on_waiting
        proc.add_process_listener(listener)

    return in_waiting


def run_until_paused(proc):
    """ Set up a future that will be resolved when the process is paused """

    listener = plumpy.ProcessListener()
    paused = plumpy.Future()

    if proc.paused:
        paused.set_result(True)
    else:

        def on_paused(_paused_proc):
            paused.set_result(True)
            proc.remove_process_listener(listener)

        listener.on_process_paused = on_paused
        proc.add_process_listener(listener)

    return paused


async def wait_util(condition, sleep_interval=0.1):
    """Given a condition function, keep polling until it returns True"""
    while not condition():
        await asyncio.sleep(sleep_interval)



# -*- coding: utf-8 -*-
__all__ = ['super_check', 'call_with_super_check']


def super_check(wrapped):
    """
    Decorator to add a super check to a function to be used with
    call_with_super_check
    """

    def wrapper(self, *args, **kwargs):
        msg = "The function '{}' was not called through call_with_super_check".format(wrapped.__name__)
        assert getattr(self, '_called', 0) >= 1, msg
        wrapped(self, *args, **kwargs)
        self._called -= 1

    return wrapper


def call_with_super_check(wrapped, *args, **kwargs):
    """
    Call a class method checking that all subclasses called super along the way
    """
    self = wrapped.__self__
    call_count = getattr(self, '_called', 0)
    self._called = call_count + 1
    wrapped(*args, **kwargs)
    msg = "Base '{}' was not called from '{}'\nHint: Did you forget to call the superclass method?".format(
        wrapped.__name__, self.__class__
    )
    assert self._called == call_count, msg



# -*- coding: utf-8 -*-
import inspect
import unittest
import asyncio

import pytest
import plumpy
from plumpy.workchains import *
from plumpy.process_listener import ProcessListener

from . import utils


class Wf(WorkChain):
    # Keep track of which steps were completed by the workflow
    finished_steps = {}

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input('value', default='A')
        spec.input('n', default=3)
        spec.outputs.dynamic = True
        spec.outline(
            cls.s1,
            if_(cls.isA)(cls.s2).elif_(cls.isB)(cls.s3).else_(cls.s4),
            cls.s5,
            while_(cls.ltN)(cls.s6),
        )

    def on_create(self):
        super().on_create()
        # Reset the finished step
        self.finished_steps = {
            k: False for k in [
                self.s1.__name__, self.s2.__name__, self.s3.__name__, self.s4.__name__, self.s5.__name__,
                self.s6.__name__, self.isA.__name__, self.isB.__name__, self.ltN.__name__
            ]
        }

    def s1(self):
        self._set_finished(inspect.stack()[0][3])

    def s2(self):
        self._set_finished(inspect.stack()[0][3])

    def s3(self):
        self._set_finished(inspect.stack()[0][3])

    def s4(self):
        self._set_finished(inspect.stack()[0][3])

    def s5(self):
        self.ctx.counter = 0
        self._set_finished(inspect.stack()[0][3])

    def s6(self):
        self.ctx.counter = self.ctx.counter + 1
        self._set_finished(inspect.stack()[0][3])

    def isA(self):
        self._set_finished(inspect.stack()[0][3])
        return self.inputs.value == 'A'

    def isB(self):
        self._set_finished(inspect.stack()[0][3])
        return self.inputs.value == 'B'

    def ltN(self):
        keep_looping = self.ctx.counter < self.inputs.n
        if not keep_looping:
            self._set_finished(inspect.stack()[0][3])
        return keep_looping

    def _set_finished(self, function_name):
        self.finished_steps[function_name] = True


class IfTest(WorkChain):

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.outline(if_(cls.condition)(cls.step1, cls.step2))

    def on_create(self, *args, **kwargs):
        super().on_create(*args, **kwargs)
        self.ctx.s1 = False
        self.ctx.s2 = False

    def condition(self):
        return True

    def step1(self):
        self.ctx.s1 = True
        self.pause()

    def step2(self):
        self.ctx.s2 = True


class DummyWc(WorkChain):

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.outline(cls.do_nothing)

    def do_nothing(self):
        pass


class TestContext(unittest.TestCase):

    def test_attributes(self):
        wc = DummyWc()
        wc.ctx.new_attr = 5
        self.assertEqual(wc.ctx.new_attr, 5)

        del wc.ctx.new_attr
        with self.assertRaises(AttributeError):
            wc.ctx.new_attr

    def test_dict(self):
        wc = DummyWc()
        wc.ctx['new_attr'] = 5
        self.assertEqual(wc.ctx['new_attr'], 5)

        del wc.ctx['new_attr']
        with self.assertRaises(KeyError):
            wc.ctx['new_attr']


class TestWorkchain(unittest.TestCase):
    maxDiff = None

    def test_run(self):
        A = 'A'
        B = 'B'
        C = 'C'
        three = 3

        # Try the if(..) part
        Wf(inputs=dict(value=A, n=three)).execute()
        # Check the steps that should have been run
        for step, finished in Wf.finished_steps.items():
            if step not in ['s3', 's4', 'isB']:
                self.assertTrue(finished, 'Step {} was not called by workflow'.format(step))

        # Try the elif(..) part
        finished_steps = Wf(inputs=dict(value=B, n=three)).execute()
        # Check the steps that should have been run
        for step, finished in finished_steps.items():
            if step not in ['isA', 's2', 's4']:
                self.assertTrue(finished, 'Step {} was not called by workflow'.format(step))

        # Try the else... part
        finished_steps = Wf(inputs=dict(value=C, n=three)).execute()
        # Check the steps that should have been run
        for step, finished in finished_steps.items():
            if step not in ['isA', 's2', 'isB', 's3']:
                self.assertTrue(finished, 'Step {} was not called by workflow'.format(step))

    def test_incorrect_outline(self):

        class Wf(WorkChain):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                # Try defining an invalid outline
                spec.outline(5)

        with self.assertRaises(TypeError):
            Wf.spec()

    def test_same_input_node(self):

        class Wf(WorkChain):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.input('a', valid_type=int)
                spec.input('b', valid_type=int)
                # Try defining an invalid outline
                spec.outline(cls.check_a_b)

            def check_a_b(self):
                assert 'a' in self.inputs
                assert 'b' in self.inputs

        x = 1
        Wf(inputs=dict(a=x, b=x)).execute()

    def test_context(self):
        A = 'a'
        B = 'b'

        class ReturnA(plumpy.Process):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.output('res')

            def run(self):
                self.out('res', A)

        class ReturnB(plumpy.Process):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.output('res')

            def run(self):
                self.out('res', B)

        class Wf(WorkChain):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.outline(cls.s1, cls.s2, cls.s3)

            def s1(self):
                return ToContext(r1=self.launch(ReturnA), r2=self.launch(ReturnB))

            def s2(self):
                assert self.ctx.r1['res'] == A
                assert self.ctx.r2['res'] == B

                # Try overwriting r1
                return ToContext(r1=self.launch(ReturnB))

            def s3(self):
                assert self.ctx.r1['res'] == B
                assert self.ctx.r2['res'] == B

        Wf().execute()

    def test_str(self):
        self.assertIsInstance(str(Wf.spec()), str)

    def test_malformed_outline(self):
        """
        Test some malformed outlines
        """
        spec = WorkChainSpec()

        with self.assertRaises(TypeError):
            spec.outline(5)

        with self.assertRaises(TypeError):
            spec.outline(lambda x, y: 5)

    def test_checkpointing(self):
        A = 'A'
        B = 'B'
        C = 'C'
        three = 3

        # Try the if(..) part
        finished_steps = self._run_with_checkpoints(Wf, inputs={'value': A, 'n': three})
        # Check the steps that should have been run
        for step, finished in finished_steps.items():
            if step not in ['s3', 's4', 'isB']:
                self.assertTrue(finished, 'Step {} was not called by workflow'.format(step))

        # Try the elif(..) part
        finished_steps = self._run_with_checkpoints(Wf, inputs={'value': B, 'n': three})
        # Check the steps that should have been run
        for step, finished in finished_steps.items():
            if step not in ['isA', 's2', 's4']:
                self.assertTrue(finished, 'Step {} was not called by workflow'.format(step))

        # Try the else... part
        finished_steps = self._run_with_checkpoints(Wf, inputs={'value': C, 'n': three})
        # Check the steps that should have been run
        for step, finished in finished_steps.items():
            if step not in ['isA', 's2', 'isB', 's3']:
                self.assertTrue(finished, 'Step {} was not called by workflow'.format(step))

    def test_return_in_outline(self):

        class WcWithReturn(WorkChain):
            FAILED_CODE = 1

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.input('success', valid_type=bool, required=False)
                spec.outline(
                    cls.step_one,
                    if_(cls.do_success)(return_).elif_(cls.do_failed)(return_(cls.FAILED_CODE)).else_(cls.default),
                )

            def step_one(self):
                pass

            def do_success(self):
                return 'success' in self.inputs and self.inputs.success is True

            def do_failed(self):
                return 'success' in self.inputs and self.inputs.success is False

            def default(self):
                raise RuntimeError('Should already have returned')

        workchain = WcWithReturn(inputs=dict(success=True))
        workchain.execute()

        workchain = WcWithReturn(inputs=dict(success=False))
        workchain.execute()

        with self.assertRaises(RuntimeError):
            workchain = WcWithReturn()
            workchain.execute()

    def test_return_in_step(self):

        class WcWithReturn(WorkChain):

            FAILED_CODE = 1

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.input('success', valid_type=bool, required=False)
                spec.outline(cls.step_one, cls.after)

            def step_one(self):
                if 'success' not in self.inputs:
                    return
                elif self.inputs.success is True:
                    return 0
                elif self.inputs.success is False:
                    return self.FAILED_CODE

            def after(self):
                raise RuntimeError('Should already have returned')

        workchain = WcWithReturn(inputs=dict(success=True))
        workchain.execute()

        workchain = WcWithReturn(inputs=dict(success=False))
        workchain.execute()

        with self.assertRaises(RuntimeError):
            workchain = WcWithReturn()
            workchain.execute()

    def test_tocontext_schedule_workchain(self):

        class MainWorkChain(WorkChain):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.outline(cls.run, cls.check)
                spec.outputs.dynamic = True

            def run(self):
                return ToContext(subwc=self.launch(SubWorkChain))

            def check(self):
                assert self.ctx.subwc.out.value == 5

        class SubWorkChain(WorkChain):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.outline(cls.run)

            def run(self):
                self.out('value', 5)

        workchain = MainWorkChain()
        workchain.execute()

    def test_if_block_persistence(self):
        workchain = IfTest()

        async def async_test():
            await utils.run_until_paused(workchain)
            self.assertTrue(workchain.ctx.s1)
            self.assertFalse(workchain.ctx.s2)

            # Now bundle the thing
            bundle = plumpy.Bundle(workchain)

            # Load from saved state
            workchain2 = bundle.unbundle()
            self.assertTrue(workchain2.ctx.s1)
            self.assertFalse(workchain2.ctx.s2)

            bundle2 = plumpy.Bundle(workchain2)
            self.assertDictEqual(bundle, bundle2)

            workchain.play()
            await workchain.future()
            self.assertTrue(workchain.ctx.s1)
            self.assertTrue(workchain.ctx.s2)

        loop = asyncio.get_event_loop()
        loop.create_task(workchain.step_until_terminated())
        loop.run_until_complete(async_test())

    def test_to_context(self):
        val = 5

        class SimpleWc(plumpy.Process):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.output('_return')

            def run(self):
                self.out('_return', val)

        class Workchain(WorkChain):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.outline(cls.begin, cls.check)

            def begin(self):
                self.to_context(result_a=self.launch(SimpleWc))
                return ToContext(result_b=self.launch(SimpleWc))

            def check(self):
                assert self.ctx.result_a['_return'] == val
                assert self.ctx.result_b['_return'] == val

        workchain = Workchain()
        workchain.execute()

    def test_output_namespace(self):
        """Test running a workchain with nested outputs."""

        class TestWorkChain(WorkChain):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.output('x.y', required=True)
                spec.outline(cls.do_run)

            def do_run(self):
                self.out('x.y', 5)

        workchain = TestWorkChain()
        workchain.execute()

    def test_exception_tocontext(self):
        my_exception = RuntimeError('Should not be reached')

        class Workchain(WorkChain):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.outline(cls.begin, cls.check)

            def begin(self):
                self.to_context(result_a=self.launch(utils.ExceptionProcess))

            def check(self):
                raise my_exception

        workchain = Workchain()
        with self.assertRaises(RuntimeError):
            workchain.execute()
        self.assertNotEqual(workchain.exception(), my_exception)

    def _run_with_checkpoints(self, wf_class, inputs=None):
        # TODO: Actually save at each point!
        proc = wf_class(inputs=inputs)
        proc.execute()
        return wf_class.finished_steps

    def test_stepper_info(self):
        """Check status information provided by steppers"""

        class Wf(WorkChain):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.input('N', valid_type=int)
                spec.outline(
                    cls.check_n,
                    while_(cls.do_step)(
                        cls.chill,
                        cls.chill,
                    ),
                    if_(cls.do_step)(cls.chill,).elif_(cls.do_step)(cls.chill,).else_(cls.chill),
                )

            def check_n(self):
                assert 'N' in self.inputs

            def chill(self):
                pass

            def do_step(self):
                if not hasattr(self.ctx, 'do_step'):
                    self.ctx.do_step = 0

                self.ctx.do_step += 1

                if self.ctx.do_step < self.inputs['N']:
                    return True
                else:
                    return False

        class StatusCollector(ProcessListener):

            def __init__(self):
                self.stepper_strings = []

            def on_process_running(self, process):
                self.stepper_strings.append(str(process._stepper))

        collector = StatusCollector()

        wf = Wf(inputs=dict(N=4))
        wf.add_process_listener(collector)
        wf.execute()

        stepper_strings = [
            '0:check_n', '1:while_(do_step)', '1:while_(do_step)(1:chill)', '1:while_(do_step)',
            '1:while_(do_step)(1:chill)', '1:while_(do_step)', '1:while_(do_step)(1:chill)', '1:while_(do_step)',
            '2:if_(do_step)'
        ]
        self.assertListEqual(collector.stepper_strings, stepper_strings)


class TestImmutableInputWorkchain(unittest.TestCase):
    """
    Test that inputs cannot be modified
    """

    def test_immutable_input(self):
        """
        Check that from within the WorkChain self.inputs returns an AttributesFrozendict which should be immutable
        """
        test_class = self

        class Wf(WorkChain):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.input('a', valid_type=int)
                spec.input('b', valid_type=int)
                spec.outline(
                    cls.step_one,
                    cls.step_two,
                )

            def step_one(self):
                # Attempt to manipulate the inputs dictionary which since it is a AttributesFrozendict should raise
                with test_class.assertRaises(TypeError):
                    self.inputs['a'] = 3
                with test_class.assertRaises(AttributeError):
                    self.inputs.pop('b')
                with test_class.assertRaises(TypeError):
                    self.inputs['c'] = 4

            def step_two(self):
                # Verify that original inputs are still there with same value and no inputs were added
                test_class.assertIn('a', self.inputs)
                test_class.assertIn('b', self.inputs)
                test_class.assertNotIn('c', self.inputs)
                test_class.assertEqual(self.inputs['a'], 1)

        workchain = Wf(inputs=dict(a=1, b=2))
        workchain.execute()

    def test_immutable_input_namespace(self):
        """
        Check that namespaced inputs also return AttributeFrozendicts and are hence immutable
        """
        test_class = self

        class Wf(WorkChain):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.input_namespace('subspace', dynamic=True)
                spec.outline(
                    cls.step_one,
                    cls.step_two,
                )

            def step_one(self):
                # Attempt to manipulate the namespaced inputs dictionary which should raise
                with test_class.assertRaises(TypeError):
                    self.inputs.subspace['one'] = 3
                with test_class.assertRaises(AttributeError):
                    self.inputs.subspace.pop('two')
                with test_class.assertRaises(TypeError):
                    self.inputs.subspace['four'] = 4

            def step_two(self):
                # Verify that original inputs are still there with same value and no inputs were added
                test_class.assertIn('one', self.inputs.subspace)
                test_class.assertIn('two', self.inputs.subspace)
                test_class.assertNotIn('four', self.inputs.subspace)
                test_class.assertEqual(self.inputs.subspace['one'], 1)

        workchain = Wf(inputs=dict(subspace={'one': 1, 'two': 2}))
        workchain.execute()



# -*- coding: utf-8 -*-
"""Module for process ports"""
import collections
import copy
import inspect
import json
import logging
import warnings

from plumpy.utils import is_mutable_property, type_check

__all__ = ['UNSPECIFIED', 'PortValidationError', 'PortNamespace', 'Port', 'InputPort', 'OutputPort']

_LOGGER = logging.getLogger(__name__)
UNSPECIFIED = ()

VALIDATOR_SIGNATURE_DEPRECATION_WARNING = """the validator `{}` has a signature that only takes a single argument.
    This has been deprecated and the new signature is `validator(value, port)` where the `port` argument will be the
    port instance to which the validator has been assigned."""


class PortValidationError(Exception):
    """Error when validation fails on a port"""

    def __init__(self, message, port):
        """
        :param message: validation error message
        :type message: str
        :param port: the port where the validation error occurred
        :type: str
        """
        super().__init__("Error occurred validating port '{}': {}".format(port, message))
        self._message = message
        self._port = port

    @property
    def message(self):
        """
        Get the validation error message

        :return: the error message
        :rtype: str
        """
        return self._message

    @property
    def port(self):
        """
        Get the port breadcrumbs

        :return: the port where the error occurred
        :rtype: str
        """
        return self._port


class Port:
    """
    Specifications relating to a general input/output value including
    properties like whether it is required, valid types, the help string, etc.
    """

    def __init__(self, name, valid_type=None, help=None, required=True, validator=None):  # pylint: disable=redefined-builtin
        self._name = name
        self._valid_type = valid_type
        self._help = help
        self._required = required
        self._validator = validator

    def __str__(self):
        """Get the string representing this port.

        :return: the string representation
        :rtype: str
        """
        return json.dumps(self.get_description())

    def get_description(self):
        """Return a description of the Port, which will be a dictionary of its attributes

        :returns: a dictionary of the stringified Port attributes
        :rtype: dict
        """
        description = {
            'name': '{}'.format(self.name),
            'required': str(self.required),
        }

        if self.valid_type:
            description['valid_type'] = '{}'.format(self.valid_type)
        if self.help:
            description['help'] = '{}'.format(self.help.strip())

        return description

    @property
    def name(self):
        return self._name

    @property
    def valid_type(self):
        """Get the valid value type for this port if one is specified

        :return: the value value type
        :rtype: type
        """
        return self._valid_type

    @valid_type.setter
    def valid_type(self, valid_type):
        """Set the valid value type for this port

        :param valid_type: the value valid type
        :type valid_type: type
        """
        self._valid_type = valid_type

    @property
    def help(self):
        """Get the help string for this port

        :return: the help string
        :rtype: str
        """
        return self._help

    @help.setter
    def help(self, help):  # pylint: disable=redefined-builtin
        """Set the help string for this port

        :param help: the help string
        :type help: str
        """
        self._help = help

    @property
    def required(self):
        """Is this port required?

        :return: True if required, False otherwise
        :rtype: bool
        """
        return self._required

    @required.setter
    def required(self, required):
        """Set whether this port is required or not

        :return: True if required, False otherwise
        :type required: bool
        """
        self._required = required

    @property
    def validator(self):
        """Get the validator for this port

        :return: the validator
        :rtype: typing.Callable[[typing.Any], typing.Tuple[bool, typing.Optional[str]]]
        """
        return self._validator

    @validator.setter
    def validator(self, validator):
        """Set the validator for this port

        :param validator: a validator function
        :type validator: typing.Callable[[typing.Any], typing.Tuple[bool, typing.Optional[str]]]
        """
        self._validator = validator

    def validate(self, value, breadcrumbs=()):
        """Validate a value to see if it is valid for this port

        :param value: the value to check
        :type value: typing.Any
        :param breadcrumbs: a tuple of the path to having reached this point in validation
        :type breadcrumbs: typing.Tuple[str]
        :return: None or tuple containing 0: error string 1: tuple of breadcrumb strings to where the validation failed
        :rtype: typing.Optional[ValidationError]
        """
        validation_error = None

        if value is UNSPECIFIED and self._required:
            validation_error = "required value was not provided for '{}'".format(self.name)
        elif value is not UNSPECIFIED and self._valid_type is not None and not isinstance(value, self._valid_type):
            validation_error = "value '{}' is not of the right type. Got '{}', expected '{}'".format(
                self.name, type(value), self._valid_type
            )

        if not validation_error and self._validator is not None and value is not UNSPECIFIED:
            spec = inspect.getfullargspec(self.validator)
            if len(spec[0]) == 1:
                warnings.warn(VALIDATOR_SIGNATURE_DEPRECATION_WARNING.format(self.validator.__name__))
                result = self.validator(value)  # pylint: disable=not-callable
            else:
                result = self.validator(value, self)  # pylint: disable=not-callable
            if result is not None:
                assert isinstance(result, str), 'Validator returned non string type'
                validation_error = result

        if validation_error:
            breadcrumbs += (self.name,)
            return PortValidationError(validation_error, breadcrumbs_to_port(breadcrumbs))


class InputPort(Port):
    """
    A simple input port for a value being received by a process
    """

    @staticmethod
    def required_override(required, default):
        """
        If a default is specified an input should no longer be marked
        as required. Otherwise the input should always be marked explicitly
        to be not required even if a default is specified.
        """
        if default is UNSPECIFIED:
            return required

        return False

    def __init__(self, name, valid_type=None, help=None, default=UNSPECIFIED, required=True, validator=None):  # pylint: disable=redefined-builtin,too-many-arguments
        super().__init__(
            name,
            valid_type=valid_type,
            help=help,
            required=InputPort.required_override(required, default),
            validator=validator
        )

        if required is not InputPort.required_override(required, default):
            _LOGGER.debug(
                "the required attribute for the input port '%s' was overridden because a default was specified", name
            )

        if default is not UNSPECIFIED:

            # Only validate the default value if it is not a callable. If it is a callable its return value will always
            # be validated when the port is validated upon process construction, if the default is was actually used.
            if not callable(default):
                validation_error = self.validate(default)
                if validation_error:
                    raise ValueError('Invalid default value: {}'.format(validation_error.message))

        self._default = default

    def has_default(self):
        return self._default is not UNSPECIFIED

    @property
    def default(self):
        if not self.has_default():
            raise RuntimeError('No default')
        return self._default

    @default.setter
    def default(self, default):
        self._default = default

    def get_description(self):
        """
        Return a description of the InputPort, which will be a dictionary of its attributes

        :returns: a dictionary of the stringified InputPort attributes
        """
        description = super().get_description()

        if self.has_default():
            description['default'] = '{}'.format(self.default)

        return description


class OutputPort(Port):
    pass


class PortNamespace(collections.abc.MutableMapping, Port):
    """
    A container for Ports. Effectively it maintains a dictionary whose members are
    either a Port or yet another PortNamespace. This allows for the nesting of ports
    """

    NAMESPACE_SEPARATOR = '.'

    def __init__(
        self,
        name=None,
        help=None,
        required=True,
        validator=None,
        valid_type=None,
        default=UNSPECIFIED,
        dynamic=False,
        populate_defaults=True
    ):  # pylint: disable=redefined-builtin,too-many-arguments
        """Construct a port namespace.

        :param name: the name of the namespace
        :param help: the help string
        :param required: boolean, if `True` the validation will fail if no value is specified for this namespace
        :param validator: an optional validator for the namespace
        :param valid_type: optional tuple of valid types in the case of a dynamic namespace. Setting this to anything
            other than `None` will automatically force `dynamic` to be set to `True`.
        :param default: default value for the port
        :param dynamic: boolean, if `True`, the namespace will accept values even when no explicit port is defined
        :param populate_defaults: boolean, when set to `False`, the populating of defaults for this namespace is skipped
            entirely, including all nested namespaces, if no explicit value is passed for this port in the parent
            namespace. As soon as a value is specified in the parent namespace for this port, even if it is empty, this
            property is ignored and the population of defaults is always performed.
        """
        super().__init__(name=name, help=help, required=required, validator=validator, valid_type=valid_type)
        self._ports = {}
        self.default = default
        self.populate_defaults = populate_defaults
        self.valid_type = valid_type

        # Do not override `dynamic` if the `valid_type` is not `None` because the setter will have set it properly
        if valid_type is None:
            self.dynamic = dynamic

    def __str__(self):
        return json.dumps(self.get_description(), sort_keys=True, indent=4)

    def __iter__(self):
        return self._ports.__iter__()

    def __len__(self):
        return len(self._ports)

    def __delitem__(self, key):
        del self._ports[key]

    def __getitem__(self, key):
        return self._ports[key]

    def __setitem__(self, key, port):
        if not isinstance(port, Port):
            raise TypeError('port needs to be an instance of Port')
        self._ports[key] = port

    @property
    def ports(self):
        return self._ports

    def has_default(self):
        return self._default is not UNSPECIFIED

    @property
    def default(self):
        return self._default

    @default.setter
    def default(self, default):
        self._default = default

    @property
    def dynamic(self):
        return self._dynamic

    @dynamic.setter
    def dynamic(self, dynamic):
        self._dynamic = dynamic

    @property
    def valid_type(self):
        return super().valid_type

    @valid_type.setter
    def valid_type(self, valid_type):
        """Set the `valid_type` for the `PortNamespace`.

        If the `valid_type` is None, the `dynamic` property will be set to `False`, in all other cases `dynamic` will be
        set to True.

        :param valid_type: a tuple or single valid type that the namespace accepts
        """
        if valid_type is not None:
            self.dynamic = True

        super(PortNamespace, self.__class__).valid_type.fset(self, valid_type)  # pylint: disable=no-member

    @property
    def populate_defaults(self):
        return self._populate_defaults

    @populate_defaults.setter
    def populate_defaults(self, populate_defaults):
        self._populate_defaults = populate_defaults

    def get_description(self):
        """
        Return a dictionary with a description of the ports this namespace contains
        Nested PortNamespaces will be properly recursed and Ports will print their properties in a list

        :returns: a dictionary of descriptions of the Ports contained within this PortNamespace
        """
        description = {
            '_attrs': {
                'default': self.default,
                'dynamic': self.dynamic,
                'valid_type': str(self.valid_type),
                'required': str(self.required),
                'help': self.help,
            }
        }

        for name, port in self._ports.items():
            description[name] = port.get_description()

        return description

    def get_port(self, name):
        """
        Retrieve a (namespaced) port from this PortNamespace. If any of the sub namespaces of the terminal
        port itself cannot be found, a ValueError will be raised

        :param name: name (potentially namespaced) of the port to retrieve
        :returns: Port
        :raises: ValueError if port or namespace does not exist
        """
        if not isinstance(name, str):
            raise ValueError('name has to be a string type, not {}'.format(type(name)))

        if not name:
            raise ValueError('name cannot be an empty string')

        namespace = name.split(self.NAMESPACE_SEPARATOR)
        port_name = namespace.pop(0)

        if port_name not in self:
            raise ValueError("port '{}' does not exist in port namespace '{}'".format(port_name, self.name))

        if namespace:
            return self[port_name].get_port(self.NAMESPACE_SEPARATOR.join(namespace))

        return self[port_name]

    def create_port_namespace(self, name, **kwargs):
        """
        Create and return a new PortNamespace in this PortNamespace. If the name is namespaced, the sub PortNamespaces
        will be created recursively, except if one of the namespaces is already occupied at any level by
        a Port in which case a ValueError will be thrown

        :param name: name (potentially namespaced) of the port to create and return
        :param kwargs: constructor arguments that will be used *only* for the construction of the terminal PortNamespace
        :returns: PortNamespace
        :raises: ValueError if any sub namespace is occupied by a non-PortNamespace port
        """
        if not isinstance(name, str):
            raise ValueError('name has to be a string type, not {}'.format(type(name)))

        if not name:
            raise ValueError('name cannot be an empty string')

        namespace = name.split(self.NAMESPACE_SEPARATOR)
        port_name = namespace.pop(0)

        if port_name in self and not isinstance(self[port_name], PortNamespace):
            raise ValueError("the name '{}' in '{}' already contains a Port".format(port_name, self.name))

        # If this is True, the (sub) port namespace does not yet exist, so we create it
        if port_name not in self:

            # If there still is a `namespace`, we create a sub namespace, *without* the constructor arguments
            if namespace:
                self[port_name] = self.__class__(port_name)

            # Otherwise it is the terminal port and we construct *with* the keyword arugments
            else:
                self[port_name] = self.__class__(port_name, **kwargs)

        if namespace:
            return self[port_name].create_port_namespace(self.NAMESPACE_SEPARATOR.join(namespace), **kwargs)

        return self[port_name]

    def absorb(self, port_namespace, exclude=None, include=None, namespace_options=None):
        """Absorb another PortNamespace instance into oneself, including all its mutable properties and ports.

        Mutable properties of self will be overwritten with those of the port namespace that is to be absorbed.
        The same goes for the ports, meaning that any ports with a key that already exists in self will
        be overwritten. The namespace_options dictionary can be used to yet override the mutable properties of
        the port namespace that is to be absorbed. The exclude and include tuples can be used to exclude or
        include certain ports and both are mutually exclusive.

        :param port_namespace: instance of PortNamespace that is to be absorbed into self
        :param exclude: list or tuple of input keys to exclude from being exposed
        :param include: list or tuple of input keys to include as exposed inputs
        :param namespace_options: a dictionary with mutable PortNamespace property values to override
        :return: list of the names of the ports that were absorbed
        """
        # pylint: disable=too-many-branches
        if not isinstance(port_namespace, PortNamespace):
            raise ValueError('port_namespace has to be an instance of PortNamespace')

        if exclude is not None and include is not None:
            raise ValueError('exclude and include are mutually exclusive')

        if exclude is not None:
            type_check(exclude, (list, tuple))
        elif include is not None:
            type_check(include, (list, tuple))

        if namespace_options is None:
            namespace_options = {}

        # Overload mutable attributes of PortNamespace unless overridden by value in namespace_options
        for attr in dir(port_namespace):
            if is_mutable_property(PortNamespace, attr):
                setattr(self, attr, namespace_options.pop(attr, getattr(port_namespace, attr)))

        if namespace_options:
            raise ValueError(
                'the namespace_options {}, is not a supported PortNamespace property'.format(
                    ', '.join(list(namespace_options.keys()))
                )
            )

        absorbed_ports = []

        for port_name, port in port_namespace.items():

            # If the current port name occurs in the exclude list, simply skip it entirely, there is no need to consider
            # any of the nested ports it might have, even if it is a port namespace
            if exclude and port_name in exclude:
                continue

            if isinstance(port, PortNamespace):

                # If the name does not appear at the start of any of the include rules we continue:
                if include and not any([rule.startswith(port_name) for rule in include]):
                    continue

                # Determine the sub exclude and include rules for this specific namespace
                sub_exclude = self.strip_namespace(port_name, self.NAMESPACE_SEPARATOR, exclude)
                sub_include = self.strip_namespace(port_name, self.NAMESPACE_SEPARATOR, include)

                # Create a new namespace at `port_name` and copy the original port namespace itself such that we keep
                # all its mutable properties, but reset its ports, since those will be taken care of by the recursive
                # absorb call that will properly consider the include and exclude rules
                self[port_name] = copy.copy(port)
                self[port_name]._ports = {}  # pylint: disable=protected-access
                self[port_name].absorb(port, sub_exclude, sub_include)
            else:
                # If include rules are specified but the port name does not appear, simply skip it
                if include and port_name not in include:
                    continue

                self[port_name] = copy.deepcopy(port)

            absorbed_ports.append(port_name)

        return absorbed_ports

    def project(self, port_values):
        """
        Project a (nested) dictionary of port values onto the port dictionary of this PortNamespace. That is
        to say, return those keys of the dictionary that are shared by this PortNamespace. If a matching key
        corresponds to another PortNamespace, this method will be called recursively, passing the sub dictionary
        belonging to that port name.

        :param port_values: a dictionary where keys are port names and values are actual input values
        """
        result = {}

        for name, value in port_values.items():
            if name in self.ports:
                if isinstance(value, PortNamespace):
                    result[name] = self[name].project(value)
                else:
                    result[name] = value

        return result

    def validate(self, port_values=None, breadcrumbs=()):  # pylint: disable=arguments-differ
        """
        Validate the namespace port itself and subsequently all the port_values it contains

        :param port_values: an arbitrarily nested dictionary of parsed port values
        :type port_values: dict
        :param breadcrumbs: a tuple of the path to having reached this point in validation
        :type breadcrumbs: typing.Tuple[str]
        :return: None or tuple containing 0: error string 1: tuple of breadcrumb strings to where the validation failed
        :rtype: typing.Optional[ValidationError]
        """
        breadcrumbs_local = breadcrumbs + (self.name,)

        if not port_values:
            port_values = {}

        if not isinstance(port_values, collections.abc.Mapping):
            message = 'specified value is of type {} which is not sub class of `Mapping`'.format(type(port_values))
            return PortValidationError(message, breadcrumbs_to_port(breadcrumbs_local))

        # Turn the port values into a normal dictionary and create a shallow copy. The `validate_ports` method to which
        # the `port_values` will be passed, will pop the values corresponding to explicit ports. This is necessary for
        # the `validate_dynamic_ports` to correctly detect any implicit ports. However, the `validator` will expect the
        # entire original input namespace to be there. Since the latter has to be called after the ports have been
        # validated, we need to create a clone of the port values here.
        port_values = dict(port_values)
        port_values_clone = port_values.copy()

        # If the namespace is not required and there are no port_values specified, consider it valid
        if not port_values and not self.required:
            return

        # In all other cases, validate all input ports explicitly specified in this port namespace
        validation_error = self.validate_ports(port_values, breadcrumbs_local)
        if validation_error:
            return validation_error

        # If any port_values remain, validate against the dynamic properties of the namespace
        validation_error = self.validate_dynamic_ports(port_values, breadcrumbs)
        if validation_error:
            return validation_error

        # Validate the validator after the ports themselves, as it most likely will rely on the port values
        if self.validator is not None:
            spec = inspect.getfullargspec(self.validator)
            if len(spec[0]) == 1:
                warnings.warn(VALIDATOR_SIGNATURE_DEPRECATION_WARNING.format(self.validator.__name__))
                message = self.validator(port_values_clone)  # pylint: disable=not-callable
            else:
                message = self.validator(port_values_clone, self)  # pylint: disable=not-callable
            if message is not None:
                assert isinstance(message, str), \
                    "Validator returned something other than None or str: '{}'".format(type(message))
                return PortValidationError(message, breadcrumbs_to_port(breadcrumbs_local))

    def pre_process(self, port_values):
        """Map port values onto the port namespace, filling in values for ports with a default.

        :param port_values: the dictionary with supplied port values
        :return: an AttributesFrozenDict with pre-processed port value mapping, complemented with port default values
        """
        from . import utils

        for name, port in self.items():

            # If the port was not specified in the inputs values and the port is a namespace with the property
            # `populate_defaults=False`, we skip the pre-processing and do not populate defaults.
            if name not in port_values and isinstance(port, PortNamespace) and not port.populate_defaults:
                continue

            if name not in port_values:

                if port.has_default():
                    default = port.default
                    if callable(default):
                        port_value = default()
                    else:
                        port_value = default

                # If a namespace containing ports, create an empty dictionary so its ports can be considered recursively
                elif isinstance(port, PortNamespace) and port.ports:
                    port_value = {}
                else:
                    continue
            else:
                port_value = port_values[name]

            if isinstance(port, PortNamespace):
                port_values[name] = port.pre_process(port_value)
            else:
                port_values[name] = port_value

        return utils.AttributesFrozendict(port_values)

    def validate_ports(self, port_values, breadcrumbs):
        """
        Validate port values with respect to the explicitly defined ports of the port namespace.
        Ports values that are matched to an actual Port will be popped from the dictionary

        :param port_values: an arbitrarily nested dictionary of parsed port values
        :type port_values: dict
        :param breadcrumbs: a tuple of breadcrumbs showing the path to to the value being validated
        :type breadcrumbs: tuple
        :return: None or tuple containing 0: error string 1: tuple of breadcrumb strings to where the validation failed
        :rtype: typing.Optional[ValidationError]
        """
        for name, port in self._ports.items():
            validation_error = port.validate(port_values.pop(name, UNSPECIFIED), breadcrumbs)
            if validation_error:
                return validation_error

    def validate_dynamic_ports(self, port_values, breadcrumbs=()):
        """
        Validate port values with respect to the dynamic properties of the port namespace. It will
        check if the namespace is actually dynamic and if all values adhere to the valid types of
        the namespace if those are specified

        :param port_values: an arbitrarily nested dictionary of parsed port values
        :type port_values: dict
        :param breadcrumbs: a tuple of the path to having reached this point in validation
        :type breadcrumbs: typing.Tuple[str]
        :return: if invalid returns a string with the reason for the validation failure, otherwise None
        :rtype: typing.Optional[str]
        """
        breadcrumbs += (self.name,)

        if port_values and not self.dynamic:
            msg = 'Unexpected ports {}, for a non dynamic namespace'.format(port_values)
            return PortValidationError(msg, breadcrumbs_to_port(breadcrumbs))

        if self.valid_type is not None:
            valid_type = self.valid_type
            for port_name, port_value in port_values.items():
                if not isinstance(port_value, valid_type):
                    msg = 'Invalid type {} for dynamic port value: expected {}'.format(type(port_value), valid_type)
                    return PortValidationError(msg, breadcrumbs_to_port(breadcrumbs + (port_name,)))

    @staticmethod
    def strip_namespace(namespace, separator, rules=None):
        """Filter given exclude/include rules staring with namespace and strip the first level.

        For example if the namespace is `base` and the rules are::

            ('base.a', 'base.sub.b','relax.base.c', 'd')

        the function will return::

            ('a', 'sub.c')

        If the rules are `None`, that is what is returned as well.

        :param namespace: the string name of the namespace
        :param separator: the namespace separator string
        :param rules: the list or tuple of exclude or include rules to strip
        :return: `None` if `rules=None` or the list of stripped rules
        """
        if rules is None:
            return rules

        stripped = []

        prefix = '{}{}'.format(namespace, separator)

        for rule in rules:
            if rule.startswith(prefix):
                stripped.append(rule[len(prefix):])

        return stripped


def breadcrumbs_to_port(breadcrumbs):
    """Convert breadcrumbs to a string representing the port

    :param breadcrumbs: a tuple of the path to the port
    :type breadcrumbs: typing.Tuple[str]
    :return: the port string
    :rtype: str
    """
    return PortNamespace.NAMESPACE_SEPARATOR.join(breadcrumbs)



# -*- coding: utf-8 -*-
import asyncio
import unittest

from kiwipy import rmq

import pytest
import plumpy

from plumpy import communications, process_comms
from test import utils

from kiwipy import rmq


class Process(plumpy.Process):

    def run(self):
        pass


class CustomObjectLoader(plumpy.DefaultObjectLoader):

    def load_object(self, identifier):
        if identifier == 'jimmy':
            return Process
        else:
            return super().load_object(identifier)

    def identify_object(self, obj):
        if isinstance(obj, Process) or issubclass(obj, Process):
            return 'jimmy'
        else:
            return super().identify_object(obj)


@pytest.mark.asyncio
async def test_continue():
    persister = plumpy.InMemoryPersister()
    load_context = plumpy.LoadSaveContext()
    launcher = plumpy.ProcessLauncher(persister=persister, load_context=load_context)

    process = utils.DummyProcess()
    pid = process.pid
    persister.save_checkpoint(process)
    del process
    process = None

    result = await launcher._continue(None, **plumpy.create_continue_body(pid)[process_comms.TASK_ARGS])
    assert result == utils.DummyProcess.EXPECTED_OUTPUTS


@pytest.mark.asyncio
async def test_loader_is_used():
    """ Make sure that the provided class loader is used by the process launcher """
    loader = CustomObjectLoader()
    proc = Process()
    persister = plumpy.InMemoryPersister(loader=loader)
    persister.save_checkpoint(proc)
    launcher = plumpy.ProcessLauncher(persister=persister, loader=loader)

    continue_task = plumpy.create_continue_body(proc.pid)
    result = await launcher._continue(None, **continue_task[process_comms.TASK_ARGS])
    assert result == utils.DummyProcess.EXPECTED_OUTPUTS



# -*- coding: utf-8 -*-
"""Module for general kiwipy communication methods"""
import asyncio
import functools

import kiwipy

from . import futures
from .utils import ensure_coroutine

__all__ = [
    'Communicator', 'RemoteException', 'DeliveryFailed', 'TaskRejected', 'plum_to_kiwi_future', 'wrap_communicator'
]

RemoteException = kiwipy.RemoteException
DeliveryFailed = kiwipy.DeliveryFailed
TaskRejected = kiwipy.TaskRejected
Communicator = kiwipy.Communicator


def plum_to_kiwi_future(plum_future):
    """
    Return a kiwi future that resolves to the outcome of the plum future

    :param plum_future: the plum future
    :type plum_future: :class:`plumpy.Future`
    :return: the kiwipy future
    :rtype: :class:`kiwipy.Future`
    """
    kiwi_future = kiwipy.Future()

    def on_done(_plum_future):
        with kiwipy.capture_exceptions(kiwi_future):
            if plum_future.cancelled():
                kiwi_future.cancel()
            else:
                result = plum_future.result()
                # Did we get another future?  In which case convert it too
                if isinstance(result, futures.Future):
                    result = plum_to_kiwi_future(result)
                kiwi_future.set_result(result)

    plum_future.add_done_callback(on_done)
    return kiwi_future


def convert_to_comm(callback, loop=None):
    """
    Take a callback function and converted it to one that will schedule a callback
    on the given even loop and return a kiwi future representing the future outcome
    of the original method.

    :param loop: the even loop to schedule the callback in
    :param callback: the function to convert
    :return: a new callback function that returns a future
    """
    coro = ensure_coroutine(callback)

    def converted(communicator, *args, **kwargs):
        msg_fn = functools.partial(coro, communicator, *args, **kwargs)
        task_future = futures.create_task(msg_fn, loop)
        return plum_to_kiwi_future(task_future)

    return converted


def wrap_communicator(communicator, loop=None):
    """
    Wrap a communicator such that all callbacks made to any subscribers are scheduled on the
    given event loop.

    If the communicator is already an equivalent communicator wrapper then it will not be
    wrapped again.

    :param communicator: the communicator to wrap
    :type communicator: :class:`kiwipy.Communicator`
    :param loop: the event loop to schedule callbacks on
    :type loop: the asyncio event loop
    :return: a communicator wrapper
    :rtype: :class:`plumpy.LoopCommunicator`
    """
    if isinstance(communicator, LoopCommunicator) and communicator.loop() is loop:
        return communicator

    return LoopCommunicator(communicator, loop)


class LoopCommunicator(kiwipy.Communicator):
    """Wrapper around a `kiwipy.Communicator` that schedules any subscriber messages on a given event loop."""

    def __init__(self, communicator, loop=None):
        """
        :param communicator: The kiwipy communicator
        :type communicator: :class:`kiwipy.Communicator`
        :param loop: The event loop to schedule callbacks on
        :type loop: The asyncio event loop
        """
        assert communicator is not None

        self._communicator = communicator
        self._loop = loop or asyncio.get_event_loop()

    def loop(self):
        return self._loop

    def add_rpc_subscriber(self, subscriber, identifier=None):
        converted = convert_to_comm(subscriber, self._loop)
        return self._communicator.add_rpc_subscriber(converted, identifier)

    def remove_rpc_subscriber(self, identifier):
        return self._communicator.remove_rpc_subscriber(identifier)

    def add_task_subscriber(self, subscriber, identifier=None):
        converted = convert_to_comm(subscriber, self._loop)
        return self._communicator.add_task_subscriber(converted, identifier)

    def remove_task_subscriber(self, identifier):
        return self._communicator.remove_task_subscriber(identifier)

    def add_broadcast_subscriber(self, subscriber, identifier=None):
        converted = convert_to_comm(subscriber, self._loop)
        return self._communicator.add_broadcast_subscriber(converted, identifier)

    def remove_broadcast_subscriber(self, identifier):
        return self._communicator.remove_broadcast_subscriber(identifier)

    def task_send(self, task, no_reply=False):
        return self._communicator.task_send(task, no_reply)

    def rpc_send(self, recipient_id, msg):
        return self._communicator.rpc_send(recipient_id, msg)

    def broadcast_send(self, body, sender=None, subject=None, correlation_id=None):
        return self._communicator.broadcast_send(body, sender, subject, correlation_id)

    def is_closed(self) -> bool:
        """Return `True` if the communicator was closed"""
        return self._communicator.is_closed()

    def close(self):
        """Close a communicator, free up all resources and do not allow any further operations"""
        self._communicator.close()



# -*- coding: utf-8 -*-
"""The state machine for processes"""
from collections.abc import Iterable
import enum
import functools
import inspect
import logging
import os
import sys

from plumpy.futures import Future
from .utils import call_with_super_check, super_check

__all__ = ['StateMachine', 'StateMachineMeta', 'event', 'TransitionFailed']

_LOGGER = logging.getLogger(__name__)


class StateMachineError(Exception):
    """Base class for state machine errors"""


class StateEntryFailed(Exception):
    """
    Failed to enter a state, can provide the next state to go to via this exception
    """

    def __init__(self, state=None, *args, **kwargs):  # pylint: disable=keyword-arg-before-vararg
        super().__init__('failed to enter state')
        self.state = state
        self.args = args
        self.kwargs = kwargs


class InvalidStateError(Exception):
    """The operation is not allowed in this state."""


class EventError(StateMachineError):

    def __init__(self, evt, msg):
        super().__init__(msg)
        self.event = evt


class TransitionFailed(Exception):
    """A state transition failed"""

    def __init__(self, initial_state, final_state=None, traceback_str=None):
        self.initial_state = initial_state
        self.final_state = final_state
        self.traceback_str = traceback_str
        super().__init__(self._format_msg())

    def _format_msg(self):
        msg = ['{} -> {}'.format(self.initial_state, self.final_state)]
        if self.traceback_str is not None:
            msg.append(self.traceback_str)
        return '\n'.join(msg)


def event(from_states='*', to_states='*'):
    if from_states != '*':
        if inspect.isclass(from_states):
            from_states = (from_states,)
        if not all(issubclass(state, State) for state in from_states):
            raise TypeError()
    if to_states != '*':
        if inspect.isclass(to_states):
            to_states = (to_states,)
        if not all(issubclass(state, State) for state in to_states):
            raise TypeError()

    def wrapper(wrapped):
        evt_label = wrapped.__name__

        @functools.wraps(wrapped)
        def transition(self, *a, **kw):
            initial = self._state

            if from_states != '*' and not any(isinstance(self._state, state) for state in from_states):
                raise EventError(evt_label, 'Event {} invalid in state {}'.format(evt_label, initial.LABEL))

            result = wrapped(self, *a, **kw)
            if not (result is False or isinstance(result, Future)):
                if to_states != '*' and not any(isinstance(self._state, state) for state in to_states):
                    if self._state == initial:
                        raise EventError(evt_label, 'Machine did not transition')

                    raise EventError(
                        evt_label, 'Event produced invalid state transition from '
                        '{} to {}'.format(initial.LABEL, self._state.LABEL)
                    )

            return result

        return transition

    if inspect.isfunction(from_states):
        return wrapper(from_states)

    return wrapper


class State:
    LABEL = None
    # A set containing the labels of states that can be entered
    # from this one
    ALLOWED = set()

    @classmethod
    def is_terminal(cls):
        return not cls.ALLOWED

    def __init__(self, state_machine):
        """
        :param state_machine: The process this state belongs to
        :type state_machine: :class:`StateMachine`
        """
        self.state_machine = state_machine
        self.in_state = False

    def __str__(self):
        return str(self.LABEL)

    @property
    def label(self):
        """ Convenience property to get the state label """
        return self.LABEL

    @super_check
    def enter(self):
        """ Entering the state """

    def execute(self):
        """
        Execute the state, performing the actions that this state is responsible
        for.  Return a state to transition to or None if finished.
        """

    @super_check
    def exit(self):
        """ Exiting the state """
        if self.is_terminal():
            raise InvalidStateError('Cannot exit a terminal state {}'.format(self.LABEL))

    def create_state(self, state_label, *args, **kwargs):
        return self.state_machine.create_state(state_label, *args, **kwargs)

    def do_enter(self):
        call_with_super_check(self.enter)
        self.in_state = True

    def do_exit(self):
        call_with_super_check(self.exit)
        self.in_state = False


class StateEventHook(enum.Enum):
    """
    Hooks that can be used to register callback at various points in the state transition
    procedure.  The callback will be passed a state instance whose meaning will differ depending
    on the hook as commented below.
    """
    ENTERING_STATE = 0  # State passed will be the state that is being entered
    ENTERED_STATE = 1  # State passed will be the last state that we entered from
    EXITING_STATE = 2  # State passed will be the next state that will be entered (or None for terminal)


class StateMachineMeta(type):

    def __call__(cls, *args, **kwargs):
        """
        Create the state machine and enter the initial state.

        :param args: Any positional arguments to be passed to the constructor
        :param kwargs: Any keyword arguments to be passed to the constructor
        :return: An instance of the state machine
        """
        inst = super().__call__(*args, **kwargs)
        inst.transition_to(inst.create_initial_state())
        call_with_super_check(inst.init)
        return inst


class StateMachine(metaclass=StateMachineMeta):
    STATES = None
    _STATES_MAP = None

    _transitioning = False
    _transition_failing = False

    @classmethod
    def get_states_map(cls):
        cls.__ensure_built()
        return cls._STATES_MAP

    @classmethod
    def get_states(cls):
        if cls.STATES is not None:
            return cls.STATES

        raise RuntimeError('States not defined')

    @classmethod
    def initial_state_label(cls):
        cls.__ensure_built()
        return cls.STATES[0].LABEL  # pylint: disable=unsubscriptable-object

    @classmethod
    def get_state_class(cls, label):
        cls.__ensure_built()
        return cls._STATES_MAP[label]  # pylint: disable=unsubscriptable-object

    @classmethod
    def __ensure_built(cls):
        try:
            # Check if it's already been built (and therefore sealed)
            if cls.__getattribute__(cls, 'sealed'):
                return
        except AttributeError:
            pass

        cls.STATES = cls.get_states()
        assert isinstance(cls.STATES, Iterable)

        # Build the states map
        cls._STATES_MAP = {}
        for state_cls in cls.STATES:  # pylint: disable=not-an-iterable
            assert issubclass(state_cls, State)
            label = state_cls.LABEL
            assert label not in cls._STATES_MAP, "Duplicate label '{}'".format(label)  # pylint: disable=unsupported-membership-test
            cls._STATES_MAP[label] = state_cls  # pylint: disable=unsupported-assignment-operation

        cls.sealed = True

    def __init__(self):
        super().__init__()
        self.__ensure_built()
        self._state = None
        self._exception_handler = None
        self.set_debug((not sys.flags.ignore_environment and bool(os.environ.get('PYTHONSMDEBUG'))))
        self._transitioning = False
        self._event_callbacks = {}

    @super_check
    def init(self):
        """ Called after entering initial state. """

    def __str__(self):
        return '<{}> ({})'.format(self.__class__.__name__, self.state)

    def create_initial_state(self):
        return self.get_state_class(self.initial_state_label())(self)

    @property
    def state(self):
        if self._state is None:
            return None
        return self._state.LABEL

    def add_state_event_callback(self, hook, callback):
        """
        Add a callback to be called on a particular state event hook.
        The callback should have form fn(state_machine, hook, state)

        :param hook: The state event hook
        :param callback: The callback function
        """
        self._event_callbacks.setdefault(hook, []).append(callback)

    def remove_state_event_callback(self, hook, callback):
        try:
            self._event_callbacks[hook].remove(callback)
        except (KeyError, ValueError):
            raise ValueError("Callback not set for hook '{}'".format(hook))

    def _fire_state_event(self, hook, state):
        for callback in self._event_callbacks.get(hook, []):
            callback(self, hook, state)

    @super_check
    def on_terminated(self):
        """ Called when a terminal state is entered """

    def transition_to(self, new_state, *args, **kwargs):
        assert not self._transitioning, \
            'Cannot call transition_to when already transitioning state'

        initial_state_label = self._state.LABEL if self._state is not None else None
        label = None
        try:
            self._transitioning = True

            # Make sure we have a state instance
            new_state = self._create_state_instance(new_state, *args, **kwargs)
            label = new_state.LABEL
            self._exit_current_state(new_state)

            try:
                self._enter_next_state(new_state)
            except StateEntryFailed as exception:
                new_state = exception.state
                # Make sure we have a state instance
                new_state = self._create_state_instance(new_state, *exception.args, **exception.kwargs)
                label = new_state.LABEL
                self._exit_current_state(new_state)
                self._enter_next_state(new_state)

            if self._state.is_terminal():
                call_with_super_check(self.on_terminated)
        except Exception:  # pylint: disable=broad-except
            self._transitioning = False
            if self._transition_failing:
                raise
            self._transition_failing = True
            self.transition_failed(initial_state_label, label, *sys.exc_info()[1:])
        finally:
            self._transition_failing = False
            self._transitioning = False

    @staticmethod
    def transition_failed(initial_state, final_state, exception, trace):
        """
        Called when a state transitions fails.  This method can be overwritten
        to change the default behaviour which is to raise the exception.

        :param exception: The transition failed exception
        :type exception: :class:`Exception`
        """
        raise exception.with_traceback(trace)

    def get_debug(self):
        return self._debug

    def set_debug(self, enabled):
        self._debug = enabled

    def create_state(self, state_label, *args, **kwargs):
        try:
            return self.get_states_map()[state_label](self, *args, **kwargs)  # pylint: disable=unsubscriptable-object
        except KeyError:
            raise ValueError('{} is not a valid state'.format(state_label))

    def _exit_current_state(self, next_state):
        """ Exit the given state """

        # If we're just being constructed we may not have a state yet to exit,
        # in which case check the new state is the initial state
        if self._state is None:
            if next_state.label != self.initial_state_label():
                raise RuntimeError("Cannot enter state '{}' as the initial state".format(next_state))
            return  # Nothing to exit

        if next_state.LABEL not in self._state.ALLOWED:
            raise RuntimeError('Cannot transition from {} to {}'.format(self._state.LABEL, next_state.label))
        self._fire_state_event(StateEventHook.EXITING_STATE, next_state)
        self._state.do_exit()

    def _enter_next_state(self, next_state):
        last_state = self._state
        self._fire_state_event(StateEventHook.ENTERING_STATE, next_state)
        # Enter the new state
        next_state.do_enter()
        self._state = next_state
        self._fire_state_event(StateEventHook.ENTERED_STATE, last_state)

    def _create_state_instance(self, state, *args, **kwargs):
        if isinstance(state, State):
            # It's already a state instance
            return state

        # OK, have to create it
        state_cls = self._ensure_state_class(state)
        return state_cls(self, *args, **kwargs)

    def _ensure_state_class(self, state):
        if inspect.isclass(state) and issubclass(state, State):
            return state

        try:
            return self.get_states_map()[state]  # pylint: disable=unsubscriptable-object
        except KeyError:
            raise ValueError('{} is not a valid state'.format(state))



# -*- coding: utf-8 -*-
import unittest
import inspect
import functools
import asyncio

from plumpy.utils import AttributesFrozendict, ensure_coroutine


class TestAttributesFrozendict(unittest.TestCase):

    def test_getitem(self):
        d = AttributesFrozendict({'a': 5})
        self.assertEqual(d['a'], 5)

        with self.assertRaises(KeyError):
            d['b']

    def test_getattr(self):
        d = AttributesFrozendict({'a': 5})
        self.assertEqual(d.a, 5)

        with self.assertRaises(AttributeError):
            d.b

    def test_setitem(self):
        d = AttributesFrozendict()
        with self.assertRaises(TypeError):
            d['a'] = 5


def fct():
    pass


async def async_fct():
    pass


class TestEnsureCoroutine(unittest.TestCase):

    def test_sync_func(self):
        coro = ensure_coroutine(fct)
        assert inspect.iscoroutinefunction(coro)

    def test_async_func(self):
        coro = ensure_coroutine(async_fct)
        assert coro is async_fct

    def test_callable_class(self):
        """
        """

        class AsyncDummy:

            async def __call__(self):
                pass

        coro = ensure_coroutine(AsyncDummy)
        assert coro is AsyncDummy

    def test_callable_object(self):
        """
        """

        class AsyncDummy:

            async def __call__(self):
                pass

        obj = AsyncDummy()
        coro = ensure_coroutine(obj)
        assert coro is obj

    def test_functools_partial(self):
        fct_wrap = functools.partial(fct)
        coro = ensure_coroutine(fct_wrap)
        assert coro is not fct_wrap
        assert asyncio.iscoroutine(coro())



# -*- coding: utf-8 -*-
from .utils import TestCase
from plumpy.lang import protected, override


class A:

    def __init__(self):
        self._a = None

    @protected(check=True)
    def protected_fn(self):
        return self._a

    @property
    @protected(check=True)
    def protected_property(self):
        return self._a

    @protected(check=False)
    def protected_fn_nocheck(self):
        return self._a

    def testA(self):
        self.protected_fn()
        self.protected_property


class B(A):

    def testB(self):
        self.protected_fn()
        self.protected_property


class C(B):

    def testC(self):
        self.protected_fn()
        self.protected_property


class TestProtected(TestCase):

    def test_free_function(self):
        with self.assertRaises(RuntimeError):

            @protected(check=False)
            def some_func():
                pass

    def test_correct_usage(self):
        # All A, B and C should be able to call the protected fn
        A().testA()
        B().testB()
        C().testC()

    def test_incorrect_usage(self):
        # I shouldn't be able to call the protected function from any of them
        a = A()
        with self.assertRaises(RuntimeError):
            a.protected_fn()
        with self.assertRaises(RuntimeError):
            a.protected_property

        b = B()
        with self.assertRaises(RuntimeError):
            b.protected_fn()
        with self.assertRaises(RuntimeError):
            b.protected_property

        c = C()
        with self.assertRaises(RuntimeError):
            c.protected_fn()
        with self.assertRaises(RuntimeError):
            c.protected_property

        with self.assertRaises(RuntimeError):

            class TestWrongDecoratorOrder:

                @protected(check=True)
                @property
                def a(self):
                    return None


class Superclass:

    def test(self):
        pass


class TestOverride(TestCase):

    def test_free_function(self):
        with self.assertRaises(RuntimeError):

            @override(check=False)
            def some_func():
                pass

    def test_correct_usage(self):

        class Derived(Superclass):

            @override(check=True)
            def test(self):
                return True

        self.assertTrue(Derived().test())

        class Middle(Superclass):
            pass

        class Next(Middle):

            @override(check=True)
            def test(self):
                return True

        self.assertTrue(Next().test())

    def test_incorrect_usage(self):

        class Derived:

            @override(check=True)
            def test(self):
                pass

        with self.assertRaises(RuntimeError):
            Derived().test()

        with self.assertRaises(RuntimeError):

            class TestWrongDecoratorOrder(Superclass):

                @override(check=True)
                @property
                def test(self):
                    return None


#
#
# class A:
#     def __init__(self):
#         self.a_called = False
#
#     def test(self):
#         self.a_called = True
#
#
# class B(A):
#     def __init__(self):
#         super().__init__()
#         self.b_called = False
#
#     @call_super
#     def test(self):
#         self.b_called = True
#
#
# class C(B):
#     def __init__(self):
#         super().__init__()
#         self.c_called = False
#
#     @call_super
#     def test(self):
#         self.c_called = True
#
# class BPrime(A):
#     def __init__(self):
#         super().__init__()
#         self.b_prime_called = False
#
#     def test(self):
#         self.b_prime_called = True
#
# class CPrime(BPrime):
#     def __init__(self):
#         super().__init__()
#         self.c_prime_called = False
#
#     @call_super
#     def test(self):
#         self._c_prime_called = True

#  class TestCallSuper(TestCase):
#     def test_one_up(self):
#         b = B()
#         b.test()
#         self.assertTrue(b.a_called)
#         self.assertTrue(b.b_called)
#
#     def test_two_up(self):
#         c = C()
#         c.test()
#         self.assertTrue(c.a_called)
#         self.assertTrue(c.b_called)
#         self.assertTrue(c.c_called)
#
#     def test_two_up_skip_one(self):
#         c_prime = CPrime()
#         c_prime.test()
#         self.assertTrue(c_prime.a_called)
#         self.assertTrue(c_prime.b_prime_called)
#         self.assertFalse(c_prime.a_called)
#



# -*- coding: utf-8 -*-
import unittest
from plumpy.base import utils


class Root:

    @utils.super_check
    def method(self):
        pass

    def do(self):
        utils.call_with_super_check(self.method)


class DoCall(Root):

    def method(self):
        super().method()


class DontCall(Root):

    def method(self):
        pass


class TestSuperCheckMixin(unittest.TestCase):

    def test_do_call(self):
        DoCall().do()

    def test_dont_call(self):
        with self.assertRaises(AssertionError):
            DontCall().do()

    def dont_call_middle(self):

        class ThirdChild(DontCall):

            def method(self):
                super().method()

        with self.assertRaises(AssertionError):
            ThirdChild.do()

    def test_skip_check_call(self):
        with self.assertRaises(AssertionError):
            DoCall().method()



# -*- coding: utf-8 -*-
from enum import Enum
import sys
import traceback

import yaml

try:
    import tblib

    _HAS_TBLIB = True
except ImportError:
    _HAS_TBLIB = False

from . import futures
from .base import state_machine
from . import persistence
from .persistence import auto_persist
from .lang import NULL
from . import utils
from . import exceptions

__all__ = [
    'ProcessState',
    'Created',
    'Running',
    'Waiting',
    'Finished',
    'Excepted',
    'Killed',
    # Commands
    'Kill',
    'Stop',
    'Wait',
    'Continue',
    'Interruption',
    'KillInterruption',
    'PauseInterruption',
]


class Interruption(Exception):
    pass


class KillInterruption(Interruption):
    pass


class PauseInterruption(Interruption):
    pass


# region Commands


class Command(persistence.Savable):
    pass


@auto_persist('msg')
class Kill(Command):

    def __init__(self, msg=None):
        super().__init__()
        self.msg = msg


class Pause(Command):
    pass


@auto_persist('msg', 'data')
class Wait(Command):

    def __init__(self, continue_fn=None, msg=None, data=None):
        super().__init__()
        self.continue_fn = continue_fn
        self.msg = msg
        self.data = data


@auto_persist('result')
class Stop(Command):

    def __init__(self, result, successful):
        super().__init__()
        self.result = result
        self.successful = successful


@auto_persist('args', 'kwargs')
class Continue(Command):
    CONTINUE_FN = 'continue_fn'

    def __init__(self, continue_fn, *args, **kwargs):
        super().__init__()
        self.continue_fn = continue_fn
        self.args = args
        self.kwargs = kwargs

    def save_instance_state(self, out_state, save_context):
        super().save_instance_state(out_state, save_context)
        out_state[self.CONTINUE_FN] = self.continue_fn.__name__

    def load_instance_state(self, saved_state, load_context):
        super().load_instance_state(saved_state, load_context)
        try:
            self.continue_fn = utils.load_function(saved_state[self.CONTINUE_FN])
        except ValueError:
            process = load_context.process
            self.continue_fn = getattr(process, saved_state[self.CONTINUE_FN])


# endregion

# region States


class ProcessState(Enum):
    """
    The possible states that a :class:`Process` can be in.
    """
    CREATED = 'created'
    RUNNING = 'running'
    WAITING = 'waiting'
    FINISHED = 'finished'
    EXCEPTED = 'excepted'
    KILLED = 'killed'


@auto_persist('in_state')
class State(state_machine.State, persistence.Savable):

    @property
    def process(self):
        """
        :return: The process
        :rtype: :class:`ProcessStateMachine`
        """
        return self.state_machine

    def load_instance_state(self, saved_state, load_context):
        super().load_instance_state(saved_state, load_context)
        self.state_machine = load_context.process

    @staticmethod
    def interrupt(reason):  # pylint: disable=unused-argument
        return False


@auto_persist('args', 'kwargs')
class Created(State):
    LABEL = ProcessState.CREATED
    ALLOWED = {ProcessState.RUNNING, ProcessState.KILLED, ProcessState.EXCEPTED}

    RUN_FN = 'run_fn'

    def __init__(self, process, run_fn, *args, **kwargs):
        super().__init__(process)
        assert run_fn is not None
        self.run_fn = run_fn
        self.args = args
        self.kwargs = kwargs

    def save_instance_state(self, out_state, save_context):
        super().save_instance_state(out_state, save_context)
        out_state[self.RUN_FN] = self.run_fn.__name__

    def load_instance_state(self, saved_state, load_context):
        super().load_instance_state(saved_state, load_context)
        self.run_fn = getattr(self.process, saved_state[self.RUN_FN])

    def execute(self):
        return self.create_state(ProcessState.RUNNING, self.run_fn, *self.args, **self.kwargs)


@auto_persist('args', 'kwargs')
class Running(State):
    LABEL = ProcessState.RUNNING
    ALLOWED = {
        ProcessState.RUNNING, ProcessState.WAITING, ProcessState.FINISHED, ProcessState.KILLED, ProcessState.EXCEPTED
    }

    RUN_FN = 'run_fn'  # The key used to store the function to run
    COMMAND = 'command'  # The key used to store an upcoming command

    # Class level defaults
    _command = None
    _running = False
    _run_handle = None

    def __init__(self, process, run_fn, *args, **kwargs):
        super().__init__(process)
        assert run_fn is not None
        self.run_fn = run_fn
        self.args = args
        self.kwargs = kwargs
        self._run_handle = None

    def save_instance_state(self, out_state, save_context):
        super().save_instance_state(out_state, save_context)
        out_state[self.RUN_FN] = self.run_fn.__name__
        if self._command is not None:
            out_state[self.COMMAND] = self._command.save()

    def load_instance_state(self, saved_state, load_context):
        super().load_instance_state(saved_state, load_context)
        self.run_fn = getattr(self.process, saved_state[self.RUN_FN])
        if self.COMMAND in saved_state:
            self._command = persistence.Savable.load(saved_state[self.COMMAND], load_context)

    def interrupt(self, reason):
        return False

    async def execute(self):  # pylint: disable=invalid-overridden-method
        if self._command is not None:
            command = self._command
        else:
            try:
                try:
                    self._running = True
                    result = self.run_fn(*self.args, **self.kwargs)
                finally:
                    self._running = False
            except Interruption:
                # Let this bubble up to the caller
                raise
            except Exception:  # pylint: disable=broad-except
                excepted = self.create_state(ProcessState.EXCEPTED, *sys.exc_info()[1:])
                return excepted
            else:
                if not isinstance(result, Command):
                    if isinstance(result, exceptions.UnsuccessfulResult):
                        result = Stop(result.result, False)
                    else:
                        # Got passed a basic return type
                        result = Stop(result, True)

                command = result

        next_state = self._action_command(command)
        return next_state

    def _action_command(self, command):
        if isinstance(command, Kill):
            return self.create_state(ProcessState.FINISHED, command.result, command.successful)
        # elif isinstance(command, Pause):
        #     self.pause()
        if isinstance(command, Stop):
            return self.create_state(ProcessState.FINISHED, command.result, command.successful)

        if isinstance(command, Wait):
            return self.create_state(ProcessState.WAITING, command.continue_fn, command.msg, command.data)

        if isinstance(command, Continue):
            return self.create_state(ProcessState.RUNNING, command.continue_fn, *command.args)

        raise ValueError('Unrecognised command')


@auto_persist('msg', 'data')
class Waiting(State):
    LABEL = ProcessState.WAITING
    ALLOWED = {
        ProcessState.RUNNING, ProcessState.WAITING, ProcessState.KILLED, ProcessState.EXCEPTED, ProcessState.FINISHED
    }

    DONE_CALLBACK = 'DONE_CALLBACK'

    _interruption = None

    def __str__(self):
        state_info = super().__str__()
        if self.msg is not None:
            state_info += ' ({})'.format(self.msg)
        return state_info

    def __init__(self, process, done_callback, msg=None, data=None):
        super().__init__(process)
        self.done_callback = done_callback
        self.msg = msg
        self.data = data
        self._waiting_future = futures.Future()

    def save_instance_state(self, out_state, save_context):
        super().save_instance_state(out_state, save_context)
        if self.done_callback is not None:
            out_state[self.DONE_CALLBACK] = self.done_callback.__name__

    def load_instance_state(self, saved_state, load_context):
        super().load_instance_state(saved_state, load_context)
        callback_name = saved_state.get(self.DONE_CALLBACK, None)
        if callback_name is not None:
            self.done_callback = getattr(self.process, callback_name)
        else:
            self.done_callback = None
        self._waiting_future = futures.Future()

    def interrupt(self, reason):
        # This will cause the future in execute() to raise the exception
        self._waiting_future.set_exception(reason)

    async def execute(self):  # pylint: disable=invalid-overridden-method
        try:
            result = await self._waiting_future
        except Interruption:
            # Deal with the interruption (by raising) but make sure our internal
            # state is back to how it was before the interruption so that we can be
            # re-executed
            self._waiting_future = futures.Future()
            raise

        if result == NULL:
            next_state = self.create_state(ProcessState.RUNNING, self.done_callback)
        else:
            next_state = self.create_state(ProcessState.RUNNING, self.done_callback, result)

        return next_state

    def resume(self, value=NULL):
        assert self._waiting_future is not None, 'Not yet waiting'
        self._waiting_future.set_result(value)


class Excepted(State):
    LABEL = ProcessState.EXCEPTED

    EXC_VALUE = 'ex_value'
    TRACEBACK = 'traceback'

    def __init__(self, process, exception, trace_back=None):
        """
        :param process: The associated process
        :param exception: The exception instance
        :param trace_back: An optional exception traceback
        """
        super().__init__(process)
        self.exception = exception
        self.traceback = trace_back

    def __str__(self):
        return '{} ({})'.format(
            super().__str__(),
            traceback.format_exception_only(type(self.exception), self.exception)[0]
        )

    def save_instance_state(self, out_state, save_context):
        super().save_instance_state(out_state, save_context)
        out_state[self.EXC_VALUE] = yaml.dump(self.exception)
        if self.traceback is not None:
            out_state[self.TRACEBACK] = ''.join(traceback.format_tb(self.traceback))

    def load_instance_state(self, saved_state, load_context):
        super().load_instance_state(saved_state, load_context)
        self.exception = yaml.load(saved_state[self.EXC_VALUE], Loader=yaml.FullLoader)
        if _HAS_TBLIB:
            try:
                self.traceback = \
                    tblib.Traceback.from_string(saved_state[self.TRACEBACK],
                                                strict=False)
            except KeyError:
                self.traceback = None
        else:
            self.traceback = None

    def get_exc_info(self):
        """
        Recreate the exc_info tuple and return it
        """
        return type(self.exception), self.exception, self.traceback


@auto_persist('result', 'successful')
class Finished(State):
    LABEL = ProcessState.FINISHED

    def __init__(self, process, result, successful):
        super().__init__(process)
        self.result = result
        self.successful = successful


@auto_persist('msg')
class Killed(State):
    LABEL = ProcessState.KILLED

    def __init__(self, process, msg):
        """
        :param process: The associated process
        :param msg: Optional kill message
        :type msg: str
        """
        super().__init__(process)
        self.msg = msg


# endregion



# -*- coding: utf-8 -*-
from plumpy.ports import PortNamespace, InputPort
from plumpy import ProcessSpec
from .utils import TestCase


class StrSubtype(str):
    pass


class TestProcessSpec(TestCase):

    def setUp(self):
        self.spec = ProcessSpec()

    def test_get_port_namespace_base(self):
        """
        Get the root, inputs and outputs port namespaces of the ProcessSpec
        """
        ports = self.spec.ports
        input_ports = self.spec.inputs
        output_ports = self.spec.outputs

        self.assertTrue(input_ports.name, self.spec.NAME_INPUTS_PORT_NAMESPACE)
        self.assertTrue(output_ports.name, self.spec.NAME_OUTPUTS_PORT_NAMESPACE)

    def test_dynamic_output(self):
        self.spec.outputs.dynamic = True
        self.spec.outputs.valid_type = str
        self.assertIsNone(self.spec.outputs.validate({'dummy': 'foo'}))
        self.assertIsNone(self.spec.outputs.validate({'dummy': StrSubtype('bar')}))
        self.assertIsNotNone(self.spec.outputs.validate({'dummy': 5}))

        # Remove dynamic output
        self.spec.outputs.dynamic = False
        self.spec.outputs.valid_type = None

        # Now add and check behaviour
        self.spec.outputs.dynamic = True
        self.spec.outputs.valid_type = str
        self.assertIsNone(self.spec.outputs.validate({'dummy': 'foo'}))
        self.assertIsNone(self.spec.outputs.validate({'dummy': StrSubtype('bar')}))
        self.assertIsNotNone(self.spec.outputs.validate({'dummy': 5}))

    def test_get_description(self):
        spec = ProcessSpec()

        # Adding an input should create some description
        spec.input('test')
        description = spec.get_description()
        self.assertNotEqual(description, {})

        # Similar with adding output
        spec = ProcessSpec()
        spec.output('test')
        description = spec.get_description()
        self.assertNotEqual(description, {})

    def test_input_namespaced(self):
        """
        Test the creation of a namespaced input port
        """
        self.spec.input('some.name.space.a', valid_type=int)

        self.assertTrue('some' in self.spec.inputs)
        self.assertTrue('name' in self.spec.inputs['some'])
        self.assertTrue('space' in self.spec.inputs['some']['name'])
        self.assertTrue('a' in self.spec.inputs['some']['name']['space'])

        self.assertTrue(isinstance(self.spec.inputs.get_port('some'), PortNamespace))
        self.assertTrue(isinstance(self.spec.inputs.get_port('some.name'), PortNamespace))
        self.assertTrue(isinstance(self.spec.inputs.get_port('some.name.space'), PortNamespace))
        self.assertTrue(isinstance(self.spec.inputs.get_port('some.name.space.a'), InputPort))

    def test_validator(self):
        """Test the port validator with default."""

        def dict_validator(dictionary, port):
            if 'key' not in dictionary or dictionary['key'] is not 'value':
                return 'Invalid dictionary'

        self.spec.input('dict', default={'key': 'value'}, validator=dict_validator)

        processed = self.spec.inputs.pre_process({})
        self.assertEqual(processed, {'dict': {'key': 'value'}})
        self.spec.inputs.validate()

        processed = self.spec.inputs.pre_process({'dict': {'key': 'value'}})
        self.assertEqual(processed, {'dict': {'key': 'value'}})
        self.spec.inputs.validate()

        self.assertIsNotNone(self.spec.inputs.validate({'dict': {'wrong_key': 'value'}}))

    def test_validate(self):
        """Test the global spec validator functionality."""

        def is_valid(inputs, port):
            if not ('a' in inputs) ^ ('b' in inputs):
                return 'Must have a OR b in inputs'
            return

        self.spec.input('a', required=False)
        self.spec.input('b', required=False)
        self.spec.inputs.validator = is_valid

        processed = self.spec.inputs.pre_process({'a': 'a'})
        self.assertEqual(processed, {'a': 'a'})
        self.spec.inputs.validate()

        processed = self.spec.inputs.pre_process({'b': 'b'})
        self.assertEqual(processed, {'b': 'b'})
        self.spec.inputs.validate()

        self.assertIsNotNone(self.spec.inputs.validate({}))
        self.assertIsNotNone(self.spec.inputs.validate({'a': 'a', 'b': 'b'}))



# -*- coding: utf-8 -*-
"""Event and loop related classes and functions"""
import sys
import asyncio

__all__ = [
    'new_event_loop', 'set_event_loop', 'get_event_loop', 'run_until_complete', 'set_event_loop_policy',
    'reset_event_loop_policy', 'PlumpyEventLoopPolicy'
]

get_event_loop = asyncio.get_event_loop  # pylint: disable=invalid-name


def set_event_loop(*args, **kwargs):
    raise NotImplementedError('this method is not implemented because `plumpy` uses a single reentrant loop')


def new_event_loop(*args, **kwargs):
    raise NotImplementedError('this method is not implemented because `plumpy` uses a single reentrant loop')


class PlumpyEventLoopPolicy(asyncio.DefaultEventLoopPolicy):
    """Custom event policy that always returns the same event loop that is made reentrant by ``nest_asyncio``."""

    _loop = None

    def get_event_loop(self):
        """Return the patched event loop."""
        import nest_asyncio

        if self._loop is None:
            self._loop = super().get_event_loop()
            nest_asyncio.apply(self._loop)

        return self._loop


def set_event_loop_policy():
    """Enable plumpy's event loop policy that will make event loop's reentrant."""
    asyncio.set_event_loop_policy(PlumpyEventLoopPolicy())
    # Need to call the following explicitly for `asyncio.get_event_loop` to start calling the method of the new policy
    # in case an loop is already active.
    asyncio.get_event_loop_policy().get_event_loop()


def reset_event_loop_policy():
    """Reset the event loop policy to the default."""
    loop = get_event_loop()

    # pylint: disable=protected-access
    cls = loop.__class__
    cls._run_once = cls._run_once_orig
    cls.run_forever = cls._run_forever_orig
    cls.run_until_complete = cls._run_until_complete_orig

    del cls._check_running
    del cls._check_runnung  # typo in Python 3.7 source
    del cls._run_once_orig
    del cls._run_forever_orig
    del cls._run_until_complete_orig
    del cls._nest_patched
    # pylint: enable=protected-access

    asyncio.set_event_loop_policy(None)


def run_until_complete(future, loop=None):
    loop = loop or get_event_loop()
    return loop.run_until_complete(future)


class ProcessCallback:
    """Object returned by callback registration methods."""

    __slots__ = ('_callback', '_args', '_kwargs', '_process', '_cancelled', '__weakref__')

    def __init__(self, process, callback, args, kwargs):
        self._process = process
        self._callback = callback
        self._args = args
        self._kwargs = kwargs
        self._cancelled = False

    def cancel(self):
        if not self._cancelled:
            self._cancelled = True
            self._done()

    def cancelled(self):
        return self._cancelled

    async def run(self):
        """Run the callback"""
        if not self._cancelled:
            try:
                await self._callback(*self._args, **self._kwargs)
            except Exception:  # pylint: disable=broad-except
                exc_info = sys.exc_info()
                self._process.callback_excepted(self._callback, exc_info[1], exc_info[2])
            finally:
                self._done()

    def _done(self):
        self._cleanup()

    def _cleanup(self):
        self._process = None
        self._callback = None
        self._args = None
        self._kwargs = None



# -*- coding: utf-8 -*-
import asyncio

import plumpy
import unittest
import yaml

from . import utils


class SaveEmpty(plumpy.Savable):
    pass


@plumpy.auto_persist('test', 'test_method')
class Save1(plumpy.Savable):

    def __init__(self):
        self.test = 'sup yp'
        self.test_method = self.m

    def m():
        pass


@plumpy.auto_persist('test')
class Save(plumpy.Savable):

    def __init__(self):
        self.test = Save1()


class TestSavable(unittest.TestCase):

    def test_empty_savable(self):
        self._save_round_trip(SaveEmpty())

    def test_auto_persist(self):
        self._save_round_trip(Save1())
        self._save_round_trip_with_loader(Save1())

    def test_auto_persist_savable(self):
        self._save_round_trip(Save())
        self._save_round_trip_with_loader(Save())

    def _save_round_trip(self, savable):
        """
        Do a round trip:
        1) Save `savables` state
        2) Recreate from the saved state
        3) Save the state of the recreated `Savable`
        4) Compare the two saved states (they should match)

        :type savable: :class:`plumpy.Savable`
        """
        saved_state1 = savable.save()
        loaded = savable.recreate_from(saved_state1)
        saved_state2 = loaded.save()
        self.assertDictEqual(saved_state1, saved_state2)

    def _save_round_trip_with_loader(self, savable):
        """
        Do a round trip:
        1) Save `savables` state
        2) Recreate from the saved state
        3) Save the state of the recreated `Savable`
        4) Compare the two saved states (they should match)

        :type savable: :class:`plumpy.Savable`
        """
        object_loader = plumpy.get_object_loader()
        saved_state1 = savable.save(plumpy.LoadSaveContext(object_loader))
        loaded = savable.recreate_from(saved_state1)
        saved_state2 = loaded.save(plumpy.LoadSaveContext(object_loader))
        saved_state3 = loaded.save()
        self.assertDictEqual(saved_state1, saved_state2)
        self.assertNotEqual(saved_state1, saved_state3)


class TestBundle(unittest.TestCase):

    def test_bundle_load_context(self):
        """ Check that the loop from the load context is used """
        loop1 = asyncio.get_event_loop()
        proc = utils.DummyProcess(loop=loop1)
        bundle = plumpy.Bundle(proc)

        loop2 = asyncio.new_event_loop()
        proc2 = bundle.unbundle(plumpy.LoadSaveContext(loop=loop2))
        self.assertIs(loop2, proc2.loop)

    def test_bundle_yaml(self):
        bundle = plumpy.Bundle(Save1())
        represent = yaml.dump({'bundle': bundle})

        bundle_loaded = yaml.load(represent, Loader=yaml.Loader)['bundle']
        self.assertIsInstance(bundle_loaded, plumpy.Bundle)
        self.assertDictEqual(bundle_loaded, Save1().save())



# -*- coding: utf-8 -*-
"""Process tests"""
import unittest
import pytest
import asyncio
import enum
import kiwipy

import plumpy
from plumpy import Process, ProcessState, BundleKeys
from plumpy.utils import AttributesFrozendict
from test import utils


class ForgetToCallParent(plumpy.Process):

    def __init__(self, forget_on):
        super().__init__()
        self.forget_on = forget_on

    def on_create(self):
        if self.forget_on != 'create':
            super().on_create()

    def on_run(self):
        if self.forget_on != 'run':
            super().on_run()

    def on_except(self, exception):
        if self.forget_on != 'except':
            super().on_except(exception)

    def on_finish(self, result, successful):
        if self.forget_on != 'finish':
            super().on_finish(result, successful)

    def on_kill(self, msg):
        if self.forget_on != 'kill':
            super().on_kill(msg)


@pytest.mark.asyncio
async def test_process_scope():

    class ProcessTaskInterleave(plumpy.Process):

        async def task(self, steps: list):
            steps.append('[{}] started'.format(self.pid))
            assert plumpy.Process.current() is self
            steps.append('[{}] sleeping'.format(self.pid))
            await asyncio.sleep(0.1)
            assert plumpy.Process.current() is self
            steps.append('[{}] finishing'.format(self.pid))

    p1 = ProcessTaskInterleave()
    p2 = ProcessTaskInterleave()

    p1steps = []
    p2steps = []
    p1task = asyncio.ensure_future(p1._run_task(p1.task, p1steps))
    p2task = asyncio.ensure_future(p2._run_task(p2.task, p2steps))
    await p1task, p2task


class TestProcess(unittest.TestCase):

    def test_spec(self):
        """
        Check that the references to specs are doing the right thing...
        """
        proc = utils.DummyProcess()
        self.assertIsNot(utils.DummyProcess.spec(), Process.spec())
        self.assertIs(proc.spec(), utils.DummyProcess.spec())

        class Proc(utils.DummyProcess):
            pass

        self.assertIsNot(Proc.spec(), Process.spec())
        self.assertIsNot(Proc.spec(), utils.DummyProcess.spec())
        p = Proc()
        self.assertIs(p.spec(), Proc.spec())

    def test_dynamic_inputs(self):

        class NoDynamic(Process):
            pass

        class WithDynamic(Process):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.inputs.dynamic = True

        with self.assertRaises(ValueError):
            NoDynamic(inputs={'a': 5}).execute()

        proc = WithDynamic(inputs={'a': 5})
        proc.execute()

    def test_inputs(self):

        class Proc(Process):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.input('a')

        p = Proc({'a': 5})

        # Check that we can access the inputs after creating
        self.assertEqual(p.raw_inputs.a, 5)
        with self.assertRaises(AttributeError):
            p.raw_inputs.b

    def test_inputs_default(self):

        class Proc(utils.DummyProcess):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.input('input', default=5, required=False)

        # Supply a value
        p = Proc(inputs={'input': 2})
        self.assertEqual(p.inputs['input'], 2)

        # Don't supply, use default
        p = Proc()
        self.assertEqual(p.inputs['input'], 5)

    def test_optional_namespace(self):
        """Process with an optional namespace should not have that in `self.inputs` if not explicitly passed."""

        class SomeProcess(Process):
            """Process with single dynamic optional namespace."""

            @classmethod
            def define(cls, spec):
                super(SomeProcess, cls).define(spec)
                spec.input_namespace('namespace', required=False, valid_type=int, dynamic=True)

        # If a value is specified for `namespace` it should be present in parsed inputs
        process = SomeProcess(inputs={'namespace': {'a': 1}})
        self.assertIn('namespace', process.inputs)
        self.assertIn('a', process.inputs['namespace'])
        self.assertEqual(process.inputs['namespace']['a'], 1)

        # If nothing is passed, it should not be present
        process = SomeProcess()
        self.assertNotIn('namespace', process.inputs)

        # However, if something is passed it should be there even if it is just an empty mapping
        process = SomeProcess(inputs={'namespace': {}})
        self.assertIn('namespace', process.inputs)

        class SomeDefaultProcess(Process):
            """Process with single dynamic optional namespace, but with one concrete port with a default."""

            @classmethod
            def define(cls, spec):
                super(SomeDefaultProcess, cls).define(spec)
                spec.input_namespace('namespace', required=False, valid_type=int, dynamic=True)
                spec.input('namespace.b', default=5)

        # Even though `namespace` is optional and it is not explicitly passed as input, because the port `b` nested
        # within it has a default, the `namespace` mapping should be present in the parsed inputs.
        process = SomeDefaultProcess()
        self.assertIn('namespace', process.inputs)
        self.assertIn('b', process.inputs['namespace'])
        self.assertEqual(process.inputs['namespace']['b'], 5)

    def test_inputs_default_that_evaluate_to_false(self):
        for def_val in (True, False, 0, 1):

            class Proc(utils.DummyProcess):

                @classmethod
                def define(cls, spec):
                    super().define(spec)
                    spec.input('input', default=def_val)

            # Don't supply, use default
            p = Proc()
            self.assertIn('input', p.inputs)
            self.assertEqual(p.inputs['input'], def_val)

    def test_nested_namespace_defaults(self):
        """Process with a default in a nested namespace should be created, even if top level namespace not supplied."""

        class SomeProcess(Process):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.input_namespace('namespace', required=False)
                spec.input('namespace.sub', default=True)

        process = SomeProcess()
        self.assertIn('sub', process.inputs.namespace)
        self.assertEqual(process.inputs.namespace.sub, True)

    def test_raise_in_define(self):
        """Process which raises in its 'define' method. Check that the spec is not set."""

        class BrokenProcess(Process):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                raise ValueError

        with self.assertRaises(ValueError):
            BrokenProcess.spec()
        # Check that the error is still raised when calling .spec()
        # a second time.
        with self.assertRaises(ValueError):
            BrokenProcess.spec()

    def test_execute(self):
        proc = utils.DummyProcessWithOutput()
        proc.execute()

        self.assertTrue(proc.done())
        self.assertEqual(proc.state, ProcessState.FINISHED)
        self.assertEqual(proc.outputs, {'default': 5})

    def test_run_from_class(self):
        # Test running through class method
        proc = utils.DummyProcessWithOutput()
        proc.execute()
        results = proc.outputs
        self.assertEqual(results['default'], 5)

    def test_forget_to_call_parent(self):
        for event in ('create', 'run', 'finish'):
            with self.assertRaises(AssertionError):
                proc = ForgetToCallParent(event)
                proc.execute()

    def test_forget_to_call_parent_kill(self):
        with self.assertRaises(AssertionError):
            proc = ForgetToCallParent('kill')
            proc.kill()
            proc.execute()

    def test_pid(self):
        # Test auto generation of pid
        process = utils.DummyProcessWithOutput()
        self.assertIsNotNone(process.pid)

        # Test using integer as pid
        process = utils.DummyProcessWithOutput(pid=5)
        self.assertEqual(process.pid, 5)

        # Test using string as pid
        process = utils.DummyProcessWithOutput(pid='a')
        self.assertEqual(process.pid, 'a')

    def test_exception(self):
        proc = utils.ExceptionProcess()
        with self.assertRaises(RuntimeError):
            proc.execute()
        self.assertEqual(proc.state, ProcessState.EXCEPTED)

    def test_get_description(self):

        class ProcWithoutSpec(Process):
            pass

        class ProcWithSpec(Process):
            """ Process with a spec and a docstring """

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.input('a', default=1)

        for proc_class in utils.TEST_PROCESSES:
            desc = proc_class.get_description()
            self.assertIsInstance(desc, dict)

        desc_with_spec = ProcWithSpec.get_description()
        desc_without_spec = ProcWithoutSpec.get_description()

        self.assertIsInstance(desc_without_spec, dict)
        self.assertTrue('spec' in desc_without_spec)
        self.assertTrue('description' not in desc_without_spec)
        self.assertIsInstance(desc_with_spec['spec'], dict)

        self.assertIsInstance(desc_with_spec, dict)
        self.assertTrue('spec' in desc_with_spec)
        self.assertTrue('description' in desc_with_spec)
        self.assertIsInstance(desc_with_spec['spec'], dict)
        self.assertIsInstance(desc_with_spec['description'], str)

    def test_logging(self):

        class LoggerTester(Process):

            def run(self, **kwargs):
                self.logger.info('Test')

        # TODO: Test giving a custom logger to see if it gets used
        proc = LoggerTester()
        proc.execute()

    def test_kill(self):
        proc = utils.DummyProcess()

        proc.kill('Farewell!')
        self.assertTrue(proc.killed())
        self.assertEqual(proc.killed_msg(), 'Farewell!')
        self.assertEqual(proc.state, ProcessState.KILLED)

    def test_wait_continue(self):
        proc = utils.WaitForSignalProcess()
        # Wait - Execute the process and wait until it is waiting

        listener = plumpy.ProcessListener()
        listener.on_process_waiting = lambda proc: proc.resume()
        proc.add_process_listener(listener)

        proc.execute()

        # Check it's done
        self.assertTrue(proc.done())
        self.assertEqual(proc.state, ProcessState.FINISHED)

    def test_exc_info(self):
        proc = utils.ExceptionProcess()
        try:
            proc.execute()
        except RuntimeError as e:
            self.assertEqual(proc.exception(), e)

    def test_run_done(self):
        proc = utils.DummyProcess()
        proc.execute()
        self.assertTrue(proc.done())

    def test_wait_pause_play_resume(self):
        """
        Test that if you pause a process that and its awaitable finishes that it
        completes correctly when played again.
        """
        loop = asyncio.get_event_loop()
        proc = utils.WaitForSignalProcess()

        async def async_test():
            await utils.run_until_waiting(proc)
            self.assertEqual(proc.state, ProcessState.WAITING)

            result = await proc.pause()
            self.assertTrue(result)
            self.assertTrue(proc.paused)

            result = proc.play()
            self.assertTrue(result)
            self.assertFalse(proc.paused)

            proc.resume()
            # Wait until the process is terminated
            await proc.future()

            # Check it's done
            self.assertTrue(proc.done())
            self.assertEqual(proc.state, ProcessState.FINISHED)

        loop.create_task(proc.step_until_terminated())
        loop.run_until_complete(async_test())

    def test_pause_play_status_messaging(self):
        """
        Test the setting of a processes' status through pause and play works correctly.

        Any process can have its status set to a given message. When pausing, a pause message can be set for the
        status, which should store the current status, which should be restored, once the process is played again.
        """
        PLAY_STATUS = 'process was played by Hans Klok'
        PAUSE_STATUS = 'process was paused by Evel Knievel'

        loop = asyncio.get_event_loop()
        proc = utils.WaitForSignalProcess()
        proc.set_status(PLAY_STATUS)

        async def async_test():
            await utils.run_until_waiting(proc)
            self.assertEqual(proc.state, ProcessState.WAITING)

            result = await proc.pause(PAUSE_STATUS)
            self.assertTrue(result)
            self.assertTrue(proc.paused)
            self.assertEqual(proc.status, PAUSE_STATUS)

            result = proc.play()
            self.assertEqual(proc.status, PLAY_STATUS)
            self.assertIsNone(proc._pre_paused_status)

            proc.resume()
            # Wait until the process is terminated
            await proc.future()

        # Check it's done
        loop.create_task(proc.step_until_terminated())
        loop.run_until_complete(async_test())

        self.assertTrue(proc.done())
        self.assertEqual(proc.state, ProcessState.FINISHED)

    def test_kill_in_run(self):

        class KillProcess(Process):
            after_kill = False

            def run(self, **kwargs):
                self.kill()
                # The following line should be executed because kill will not
                # interrupt execution of a method call in the RUNNING state
                self.after_kill = True

        proc = KillProcess()
        with self.assertRaises(plumpy.KilledError):
            proc.execute()

        self.assertTrue(proc.after_kill)
        self.assertEqual(proc.state, ProcessState.KILLED)

    def test_kill_when_paused_in_run(self):

        class PauseProcess(Process):

            def run(self, **kwargs):
                self.pause()
                self.kill()

        proc = PauseProcess()
        with self.assertRaises(plumpy.KilledError):
            proc.execute()

        self.assertEqual(proc.state, ProcessState.KILLED)

    def test_kill_when_paused(self):
        loop = asyncio.get_event_loop()
        proc = utils.WaitForSignalProcess()

        async def async_test():
            await utils.run_until_waiting(proc)

            saved_state = plumpy.Bundle(proc)

            result = await proc.pause()
            self.assertTrue(result)
            self.assertTrue(proc.paused)

            # Kill the process
            proc.kill()

            with self.assertRaises(plumpy.KilledError):
                result = await proc.future()

        loop.create_task(proc.step_until_terminated())
        loop.run_until_complete(async_test())

        self.assertEqual(proc.state, ProcessState.KILLED)

    def test_run_multiple(self):
        # Create and play some processes
        loop = asyncio.get_event_loop()

        procs = []
        for proc_class in utils.TEST_PROCESSES:
            proc = proc_class()
            loop.create_task(proc.step_until_terminated())
            procs.append(proc)

        tasks = asyncio.gather(*[p.future() for p in procs])
        results = loop.run_until_complete(tasks)

        for result, proc_class in zip(results, utils.TEST_PROCESSES):
            self.assertDictEqual(proc_class.EXPECTED_OUTPUTS, result)

    def test_invalid_output(self):

        class InvalidOutput(plumpy.Process):

            def run(self):
                self.out('invalid', 5)

        proc = InvalidOutput()
        with self.assertRaises(ValueError):
            proc.execute()

    def test_missing_output(self):
        proc = utils.MissingOutputProcess()

        with self.assertRaises(plumpy.InvalidStateError):
            proc.successful()

        proc.execute()

        self.assertFalse(proc.successful())

    def test_unsuccessful_result(self):
        ERROR_CODE = 256

        class Proc(Process):

            @classmethod
            def define(cls, spec):
                super().define(spec)

            def run(self):
                return plumpy.UnsuccessfulResult(ERROR_CODE)

        proc = Proc()
        proc.execute()

        self.assertEqual(proc.result(), ERROR_CODE)

    def test_pause_in_process(self):
        """ Test that we can pause and cancel that by playing within the process """
        test_case = self

        class TestPausePlay(plumpy.Process):

            def run(self):
                fut = self.pause()
                test_case.assertIsInstance(fut, plumpy.Future)

        loop = asyncio.get_event_loop()

        listener = plumpy.ProcessListener()
        listener.on_process_paused = lambda _proc: loop.stop()

        proc = TestPausePlay()
        proc.add_process_listener(listener)

        loop.create_task(proc.step_until_terminated())
        loop.run_forever()

        self.assertTrue(proc.paused)
        self.assertEqual(plumpy.ProcessState.FINISHED, proc.state)

    def test_pause_play_in_process(self):
        """ Test that we can pause and play that by playing within the process """

        test_case = self

        class TestPausePlay(plumpy.Process):

            def run(self):
                fut = self.pause()
                test_case.assertIsInstance(fut, plumpy.Future)
                result = self.play()
                test_case.assertTrue(result)

        proc = TestPausePlay()

        proc.execute()
        self.assertFalse(proc.paused)
        self.assertEqual(plumpy.ProcessState.FINISHED, proc.state)

    def test_process_stack(self):
        test_case = self

        class StackTest(plumpy.Process):

            def run(self):
                test_case.assertIs(self, Process.current())

        proc = StackTest()
        proc.execute()

    def test_process_stack_multiple(self):
        """
        Run multiple and nested processes to make sure the process stack is always correct
        """
        expect_true = []

        def test_nested(process):
            expect_true.append(process == Process.current())

        class StackTest(plumpy.Process):

            def run(self):
                # TODO: unexpected behaviour here
                # if assert error happend here not raise
                # it will be handled by try except clause in process
                # is there better way to handle this?
                expect_true.append(self == Process.current())
                test_nested(self)

        class ParentProcess(plumpy.Process):

            def run(self):
                expect_true.append(self == Process.current())
                StackTest().execute()

        to_run = []
        n_run = 3
        for _ in range(n_run):
            to_run.append(ParentProcess().step_until_terminated())

        loop = asyncio.get_event_loop()
        loop.run_until_complete(asyncio.gather(*to_run))

        assert all(expect_true)

        self.assertEqual(len(expect_true), n_run * 3)

    def test_process_nested(self):
        """
        Run multiple and nested processes to make sure the process stack is always correct
        """

        class StackTest(plumpy.Process):

            def run(self):
                pass

        class ParentProcess(plumpy.Process):

            def run(self):
                StackTest().execute()

        ParentProcess().execute()

    def test_call_soon(self):

        class CallSoon(plumpy.Process):

            def run(self):
                self.call_soon(self.do_except)

            def do_except(self):
                raise RuntimeError('Breaking yo!')

        CallSoon().execute()

    def test_execute_twice(self):
        """Test a process that is executed once finished raises a ClosedError"""
        proc = utils.DummyProcess()
        proc.execute()
        with self.assertRaises(plumpy.ClosedError):
            proc.execute()


@plumpy.auto_persist('steps_ran')
class SavePauseProc(plumpy.Process):
    steps_ran = None

    def init(self):
        super().init()
        self.steps_ran = []

    def run(self):
        self.pause()
        self.steps_ran.append(self.run.__name__)
        return plumpy.Continue(self.step2)

    def step2(self):
        self.steps_ran.append(self.step2.__name__)


class TestProcessSaving(unittest.TestCase):
    maxDiff = None

    def test_running_save_instance_state(self):
        loop = asyncio.get_event_loop()
        nsync_comeback = SavePauseProc()

        async def async_test():
            await utils.run_until_paused(nsync_comeback)

            # Create a checkpoint
            bundle = plumpy.Bundle(nsync_comeback)
            self.assertListEqual([SavePauseProc.run.__name__], nsync_comeback.steps_ran)

            nsync_comeback.play()
            await nsync_comeback.future()

            self.assertListEqual([SavePauseProc.run.__name__, SavePauseProc.step2.__name__], nsync_comeback.steps_ran)

            proc_unbundled = bundle.unbundle()

            # At bundle time the Process was paused, the future of which will be persisted to the bundle.
            # As a result the process, recreated from that bundle, will also be paused and will have to be played
            proc_unbundled.play()
            self.assertEqual(0, len(proc_unbundled.steps_ran))
            await proc_unbundled.step_until_terminated()
            self.assertEqual([SavePauseProc.step2.__name__], proc_unbundled.steps_ran)

        loop.create_task(nsync_comeback.step_until_terminated())
        loop.run_until_complete(async_test())

    def test_save_future(self):
        """
        test `SavableFuture` is initialized with the event loop of process
        """
        loop = asyncio.new_event_loop()
        nsync_comeback = SavePauseProc(loop=loop)

        bundle = plumpy.Bundle(nsync_comeback)
        # if loop is not specified will use the default asyncio loop
        proc_unbundled = bundle.unbundle(plumpy.LoadSaveContext(loop=loop))

        async def async_test():
            await utils.run_until_paused(proc_unbundled)

            # here the future should be a SavableFuture in process loop
            proc_unbundled.play()
            await proc_unbundled.future()

            self.assertListEqual([SavePauseProc.run.__name__, SavePauseProc.step2.__name__], proc_unbundled.steps_ran)

        loop.create_task(proc_unbundled.step_until_terminated())
        loop.run_until_complete(async_test())

    def test_created_bundle(self):
        """
        Check that the bundle after just creating a process is as we expect
        """
        self._check_round_trip(utils.DummyProcess())

    def test_instance_state_with_outputs(self):
        proc = utils.DummyProcessWithOutput()

        saver = utils.ProcessSaver(proc)
        proc.execute()

        self._check_round_trip(proc)

        for bundle, outputs in zip(saver.snapshots, saver.outputs):
            # Check that it is a copy
            self.assertIsNot(outputs, bundle.get(BundleKeys.OUTPUTS, {}))
            # Check the contents are the same
            self.assertDictEqual(outputs, bundle.get(BundleKeys.OUTPUTS, {}))

        self.assertIsNot(proc.outputs, saver.snapshots[-1].get(BundleKeys.OUTPUTS, {}))

    def test_saving_each_step(self):
        loop = asyncio.get_event_loop()
        for proc_class in utils.TEST_PROCESSES:
            proc = proc_class()
            saver = utils.ProcessSaver(proc)
            saver.capture()
            self.assertEqual(proc.state, ProcessState.FINISHED)
            self.assertTrue(utils.check_process_against_snapshots(loop, proc_class, saver.snapshots))

    def test_restart(self):
        loop = asyncio.get_event_loop()
        proc = _RestartProcess()

        async def async_test():
            await utils.run_until_waiting(proc)

            # Save the state of the process
            saved_state = plumpy.Bundle(proc)

            # Load a process from the saved state
            loaded_proc = saved_state.unbundle()
            self.assertEqual(loaded_proc.state, ProcessState.WAITING)

            # Now resume it
            loaded_proc.resume()
            await loaded_proc.step_until_terminated()
            self.assertEqual(loaded_proc.outputs, {'finished': True})

        loop.create_task(proc.step_until_terminated())
        loop.run_until_complete(async_test())

    def test_wait_save_continue(self):
        """ Test that process saved while in WAITING state restarts correctly when loaded """
        loop = asyncio.get_event_loop()
        proc = utils.WaitForSignalProcess()

        async def async_test():
            await utils.run_until_waiting(proc)

            saved_state = plumpy.Bundle(proc)

            # Run the process to the end
            proc.resume()
            result1 = await proc.future()

            # Load from saved state and run again
            loader = plumpy.get_object_loader()
            proc2 = saved_state.unbundle(plumpy.LoadSaveContext(loader))
            asyncio.ensure_future(proc2.step_until_terminated())
            proc2.resume()
            result2 = await proc2.future()

            # Check results match
            self.assertEqual(result1, result2)

        loop.create_task(proc.step_until_terminated())
        loop.run_until_complete(async_test())

    def test_killed(self):
        proc = utils.DummyProcess()
        proc.kill()
        self.assertEqual(proc.state, plumpy.ProcessState.KILLED)
        self._check_round_trip(proc)

    def _check_round_trip(self, proc1):
        bundle1 = plumpy.Bundle(proc1)

        proc2 = bundle1.unbundle()
        bundle2 = plumpy.Bundle(proc2)

        self.assertEqual(proc1.pid, proc2.pid)
        self.assertDictEqual(bundle1, bundle2)


class TestProcessNamespace(unittest.TestCase):

    def test_namespaced_process(self):
        """
        Test that inputs in nested namespaces are properly validated and the returned
        Process inputs data structure consists of nested AttributesFrozenDict instances
        """

        class NameSpacedProcess(Process):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.input('some.name.space.a', valid_type=int)

        proc = NameSpacedProcess(inputs={'some': {'name': {'space': {'a': 5}}}})

        # Test that the namespaced inputs are AttributesFrozendict
        self.assertIsInstance(proc.inputs, AttributesFrozendict)
        self.assertIsInstance(proc.inputs.some, AttributesFrozendict)
        self.assertIsInstance(proc.inputs.some.name, AttributesFrozendict)
        self.assertIsInstance(proc.inputs.some.name.space, AttributesFrozendict)

        # Test that the input node is in the inputs of the process
        input_value = proc.inputs.some.name.space.a
        self.assertTrue(isinstance(input_value, int))
        self.assertEqual(input_value, 5)

    def test_namespaced_process_inputs(self):
        """
        Test the parsed inputs for a process with namespace only contains expected dictionaries
        """

        class NameSpacedProcess(Process):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.input('some.name.space.a', valid_type=int)
                spec.input('test', valid_type=int, default=6)
                spec.input('label', valid_type=str, required=False)
                spec.input('description', valid_type=str, required=False)
                spec.input('store_provenance', valid_type=bool, default=True)

        proc = NameSpacedProcess(inputs={'some': {'name': {'space': {'a': 5}}}})

        self.assertEqual(proc.inputs.test, 6)
        self.assertEqual(proc.inputs.store_provenance, True)
        self.assertEqual(proc.inputs.some.name.space.a, 5)

        self.assertTrue('label' not in proc.inputs)
        self.assertTrue('description' not in proc.inputs)

    def test_namespaced_process_dynamic(self):
        """
        Test that the input creation for processes with a dynamic nested port namespace is properly handled
        """
        namespace = 'name.space'

        class DummyDynamicProcess(Process):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.input_namespace(namespace)
                spec.inputs['name']['space'].dynamic = True
                spec.inputs['name']['space'].valid_type = int

        original_inputs = [1, 2, 3, 4]

        inputs = {'name': {'space': {str(l): l for l in original_inputs}}}
        proc = DummyDynamicProcess(inputs=inputs)

        for label, value in proc.inputs['name']['space'].items():
            self.assertTrue(label in inputs['name']['space'])
            self.assertEqual(int(label), value)
            original_inputs.remove(value)

        # Make sure there are no other inputs
        self.assertFalse(original_inputs)

    def test_namespaced_process_outputs(self):
        """Test the output namespacing and validation."""
        namespace = 'integer.namespace'

        class OutputMode(enum.Enum):

            NONE = 0
            DYNAMIC_PORT_NAMESPACE = 1
            SINGLE_REQUIRED_PORT = 2
            BOTH_SINGLE_AND_NAMESPACE = 3

        class DummyDynamicProcess(Process):

            @classmethod
            def define(cls, spec):
                super().define(spec)
                spec.input('output_mode', valid_type=OutputMode, default=OutputMode.NONE)
                spec.output('required_bool', valid_type=bool)
                spec.output_namespace(namespace, valid_type=int, dynamic=True)

            def run(self):
                if self.inputs.output_mode == OutputMode.NONE:
                    pass
                elif self.inputs.output_mode == OutputMode.DYNAMIC_PORT_NAMESPACE:
                    self.out(namespace + '.one', 1)
                    self.out(namespace + '.two', 2)
                elif self.inputs.output_mode == OutputMode.SINGLE_REQUIRED_PORT:
                    self.out('required_bool', False)
                elif self.inputs.output_mode == OutputMode.BOTH_SINGLE_AND_NAMESPACE:
                    self.out('required_bool', False)
                    self.out(namespace + '.one', 1)
                    self.out(namespace + '.two', 2)

        # Run the process in default mode which should not add any outputs and therefore fail
        process = DummyDynamicProcess()
        process.execute()

        self.assertEqual(process.state, ProcessState.FINISHED)
        self.assertFalse(process.is_successful)
        self.assertDictEqual(process.outputs, {})

        # Attaching only namespaced ports should fail, because the required port is not added
        process = DummyDynamicProcess(inputs={'output_mode': OutputMode.DYNAMIC_PORT_NAMESPACE})
        process.execute()

        self.assertEqual(process.state, ProcessState.FINISHED)
        self.assertFalse(process.is_successful)
        self.assertEqual(process.outputs['integer']['namespace']['one'], 1)
        self.assertEqual(process.outputs['integer']['namespace']['two'], 2)

        # Attaching only the single required top-level port should be fine
        process = DummyDynamicProcess(inputs={'output_mode': OutputMode.SINGLE_REQUIRED_PORT})
        process.execute()

        self.assertEqual(process.state, ProcessState.FINISHED)
        self.assertTrue(process.is_successful)
        self.assertEqual(process.outputs['required_bool'], False)

        # Attaching both the required and namespaced ports should result in a successful termination
        process = DummyDynamicProcess(inputs={'output_mode': OutputMode.BOTH_SINGLE_AND_NAMESPACE})
        process.execute()

        self.assertIsNotNone(process.outputs)
        self.assertEqual(process.state, ProcessState.FINISHED)
        self.assertTrue(process.is_successful)
        self.assertEqual(process.outputs['required_bool'], False)
        self.assertEqual(process.outputs['integer']['namespace']['one'], 1)
        self.assertEqual(process.outputs['integer']['namespace']['two'], 2)


class TestProcessEvents(unittest.TestCase):

    def test_basic_events(self):
        proc = utils.DummyProcessWithOutput()
        events_tester = utils.ProcessListenerTester(
            process=proc, expected_events=('running', 'output_emitted', 'finished')
        )
        proc.execute()
        self.assertSetEqual(events_tester.called, events_tester.expected_events)

    def test_killed(self):
        proc = utils.DummyProcessWithOutput()
        events_tester = utils.ProcessListenerTester(proc, ('killed',))
        self.assertTrue(proc.kill())

        # Do the checks
        self.assertTrue(proc.killed())
        self.assertSetEqual(events_tester.called, events_tester.expected_events)

    def test_excepted(self):
        proc = utils.ExceptionProcess()
        events_tester = utils.ProcessListenerTester(proc, (
            'excepted',
            'running',
            'output_emitted',
        ))
        with self.assertRaises(RuntimeError):
            proc.execute()
            proc.result()

        # Do the checks
        self.assertIsNotNone(proc.exception())
        self.assertSetEqual(events_tester.called, events_tester.expected_events)

    def test_paused(self):
        proc = utils.DummyProcessWithOutput()
        events_tester = utils.ProcessListenerTester(proc, ('paused',))
        self.assertTrue(proc.pause())

        # Do the checks
        self.assertSetEqual(events_tester.called, events_tester.expected_events)

    def test_broadcast(self):
        communicator = kiwipy.LocalCommunicator()

        messages = []

        def on_broadcast_receive(_comm, body, sender, subject, correlation_id):
            messages.append({'body': body, 'subject': subject, 'sender': sender, 'correlation_id': correlation_id})

        communicator.add_broadcast_subscriber(on_broadcast_receive)
        proc = utils.DummyProcess(communicator=communicator)
        proc.execute()

        expected_subjects = []
        for i, state in enumerate(utils.DummyProcess.EXPECTED_STATE_SEQUENCE):
            from_state = utils.DummyProcess.EXPECTED_STATE_SEQUENCE[i - 1].value if i != 0 else None
            expected_subjects.append('state_changed.{}.{}'.format(from_state, state.value))

        for i, message in enumerate(messages):
            self.assertEqual(message['subject'], expected_subjects[i])


class _RestartProcess(utils.WaitForSignalProcess):

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.outputs.dynamic = True

    def last_step(self):
        self.out('finished', True)



# -*- coding: utf-8 -*-
# pylint: disable=undefined-variable
import logging

from .loaders import *
from .communications import *
from .events import *
from .exceptions import *
from .futures import *
from .persistence import *
from .ports import *
from .processes import *
from .process_states import *
from .process_comms import *
from .process_listener import *
from .mixins import *
from .utils import *
from .version import *
from .workchains import *

__all__ = (
    events.__all__ + exceptions.__all__ + processes.__all__ + utils.__all__ + futures.__all__ + mixins.__all__ +
    persistence.__all__ + communications.__all__ + process_comms.__all__ + version.__all__ + process_listener.__all__ +
    workchains.__all__ + loaders.__all__ + ports.__all__ + process_states.__all__
)


# Do this se we don't get the "No handlers could be found..." warnings that will be produced
# if a user of this library doesn't set any handlers. See
# https://docs.python.org/3.1/library/logging.html#library-config
# for more details
class NullHandler(logging.Handler):

    def emit(self, record):
        pass


logging.getLogger('plumpy').addHandler(NullHandler())
